from unittest.mock import Mock

import pytest
from qtpy.QtCore import QPoint, Qt
from qtpy.QtWidgets import QApplication

from napari._app_model._app import get_app_model
from napari._qt.dialogs.qt_modal import QtPopup
from napari._qt.widgets.qt_viewer_buttons import QtViewerButtons
from napari.components.viewer_model import ViewerModel
from napari.utils.camera_orientations import (
    DepthAxisOrientation,
    HorizontalAxisOrientation,
    VerticalAxisOrientation,
)
from napari.viewer import Viewer


@pytest.fixture
def qt_viewer_buttons(qtbot):
    # create viewer model and buttons
    viewer = ViewerModel()
    viewer_buttons = QtViewerButtons(viewer)
    qtbot.addWidget(viewer_buttons)

    yield viewer, viewer_buttons

    # close still open popup widgets
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QtPopup):
            widget.close()
    viewer_buttons.close()


def test_roll_dims_button_popup(qt_viewer_buttons, qtbot):
    """
    Make sure the QtViewerButtons.rollDimsButton popup works.
    """
    # get viewer model and buttons
    viewer, viewer_buttons = qt_viewer_buttons
    assert viewer_buttons.rollDimsButton

    # make dims order settings popup
    viewer_buttons.rollDimsButton.customContextMenuRequested.emit(QPoint())

    # check that the popup widget is available
    dims_sorter_popup = None
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QtPopup):
            dims_sorter_popup = widget
    assert dims_sorter_popup


def test_grid_view_button_popup(qt_viewer_buttons, qtbot):
    """
    Make sure the QtViewerButtons.gridViewbutton popup works.

    The popup widget should be able to show/change viewer grid settings.
    """
    # get viewer model and buttons
    viewer, viewer_buttons = qt_viewer_buttons
    assert viewer_buttons.gridViewButton

    # make grid settings popup
    viewer_buttons.gridViewButton.customContextMenuRequested.emit(QPoint())

    # check popup widgets were created
    assert viewer_buttons.grid_stride_box
    assert viewer_buttons.grid_stride_box.value() == viewer.grid.stride
    assert viewer_buttons.grid_width_box
    assert viewer_buttons.grid_width_box.value() == viewer.grid.shape[1]
    assert viewer_buttons.grid_height_box
    assert viewer_buttons.grid_height_box.value() == viewer.grid.shape[0]
    assert viewer_buttons.grid_spacing_box
    assert viewer_buttons.grid_spacing_box.value() == viewer.grid.spacing

    # check that widget controls value changes update viewer grid values
    viewer_buttons.grid_stride_box.setValue(2)
    assert viewer_buttons.grid_stride_box.value() == viewer.grid.stride
    viewer_buttons.grid_width_box.setValue(2)
    assert viewer_buttons.grid_width_box.value() == viewer.grid.shape[1]
    viewer_buttons.grid_height_box.setValue(2)
    assert viewer_buttons.grid_height_box.value() == viewer.grid.shape[0]
    viewer_buttons.grid_spacing_box.setValue(0.5)
    assert viewer_buttons.grid_spacing_box.value() == viewer.grid.spacing

    # check viewer grid values changes update popup widget controls values
    viewer.grid.stride = 1
    viewer.grid.shape = (-1, -1)
    viewer.grid.spacing = 0.1
    # popup needs to be relaunched to get widget controls with the new values
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QtPopup):
            widget.close()
    viewer_buttons.gridViewButton.customContextMenuRequested.emit(QPoint())
    assert viewer_buttons.grid_stride_box.value() == viewer.grid.stride
    viewer_buttons.grid_width_box.setValue(2)
    assert viewer_buttons.grid_width_box.value() == viewer.grid.shape[1]
    assert viewer_buttons.grid_height_box.value() == viewer.grid.shape[0]
    assert viewer_buttons.grid_spacing_box.value() == viewer.grid.spacing


def test_ndisplay_button_popup(qt_viewer_buttons, qtbot):
    """
    Make sure the QtViewerButtons.ndisplayButton popup works.
    """
    # get viewer model and buttons
    viewer, viewer_buttons = qt_viewer_buttons
    assert viewer_buttons.ndisplayButton

    # trigger popup for ndisplay==2
    viewer.dims.ndisplay = 2
    viewer_buttons.ndisplayButton.customContextMenuRequested.emit(QPoint())
    perspective_popup = None
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QtPopup):
            perspective_popup = widget
    assert perspective_popup

    # check camera orientation combo changes affect viewer camera orientation
    assert viewer_buttons.vertical_combo
    assert (
        viewer_buttons.vertical_combo.currentEnum()
        == viewer.camera.orientation[1]
    )
    assert viewer_buttons.horizontal_combo
    assert (
        viewer_buttons.horizontal_combo.currentEnum()
        == viewer.camera.orientation[2]
    )
    assert viewer.camera.handedness.value == 'right'

    viewer_buttons.vertical_combo.setCurrentEnum(VerticalAxisOrientation.UP)
    viewer_buttons.horizontal_combo.setCurrentEnum(
        HorizontalAxisOrientation.LEFT
    )
    assert viewer.camera.orientation[1].value == 'up'
    assert viewer.camera.orientation[2].value == 'left'
    assert viewer.camera.orientation2d[0].value == 'up'
    assert viewer.camera.orientation2d[1].value == 'left'

    # check zoom slider change affects viewer camera zoom
    assert viewer_buttons.zoom
    viewer_buttons.zoom.setValue(3)
    assert viewer.camera.zoom == viewer_buttons.zoom.value() == 3

    # toggle ndisplay to be able to trigger ndisplay==3 popup
    viewer.dims.ndisplay = 2 + (viewer.dims.ndisplay == 2)

    # make ndisplay perspective setting popup
    viewer_buttons.ndisplayButton.customContextMenuRequested.emit(QPoint())
    perspective_popup = None
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QtPopup):
            perspective_popup = widget
    assert perspective_popup

    # check camera orientation combo changes affect viewer camera orientation
    assert viewer_buttons.depth_combo
    assert (
        viewer_buttons.depth_combo.currentEnum()
        == viewer.camera.orientation[0]
    )
    assert viewer_buttons.vertical_combo
    assert (
        viewer_buttons.vertical_combo.currentEnum()
        == viewer.camera.orientation[1]
    )
    assert viewer_buttons.horizontal_combo
    assert (
        viewer_buttons.horizontal_combo.currentEnum()
        == viewer.camera.orientation[2]
    )
    # check the values set by testing 2D orientation popup are inherited
    assert (
        viewer_buttons.vertical_combo.currentEnum()
        == VerticalAxisOrientation.UP
    )
    assert (
        viewer_buttons.horizontal_combo.currentEnum()
        == HorizontalAxisOrientation.LEFT
    )
    assert viewer.camera.handedness.value == 'right'
    assert (
        viewer_buttons.orientation_help_symbol.objectName()
        == 'righthand_label'
    )
    assert 'right-handed' in viewer_buttons.orientation_help_symbol.toolTip()

    viewer_buttons.depth_combo.setCurrentEnum(DepthAxisOrientation.AWAY)
    viewer_buttons.vertical_combo.setCurrentEnum(VerticalAxisOrientation.DOWN)
    viewer_buttons.horizontal_combo.setCurrentEnum(
        HorizontalAxisOrientation.RIGHT
    )
    assert viewer.camera.handedness.value == 'left'
    assert (
        viewer_buttons.orientation_help_symbol.objectName() == 'lefthand_label'
    )
    assert 'left-handed' in viewer_buttons.orientation_help_symbol.toolTip()
    assert viewer.camera.orientation[0].value == 'away'
    assert viewer.camera.orientation[1].value == 'down'
    assert viewer.camera.orientation[2].value == 'right'

    # check perspective slider change affects viewer camera perspective
    assert viewer_buttons.perspective
    viewer_buttons.perspective.setValue(5)
    assert viewer.camera.perspective == viewer_buttons.perspective.value() == 5

    # check zoom slider change affects viewer camera zoom
    assert viewer_buttons.zoom
    viewer_buttons.zoom.setValue(5)
    assert viewer.camera.zoom == viewer_buttons.zoom.value() == 5

    # check viewer camera rotation value affects camera angles
    assert viewer_buttons.rx
    assert viewer_buttons.ry
    assert viewer_buttons.rz
    viewer_buttons.rx.setValue(90)
    viewer_buttons.ry.setValue(45)
    viewer_buttons.rz.setValue(0)
    assert (
        viewer.camera.angles
        == (
            viewer_buttons.rx.value(),
            viewer_buttons.ry.value(),
            viewer_buttons.rz.value(),
        )
        == (90, 45, 0)
    )

    # popup needs to be relaunched to get widget controls with the new values
    perspective_popup.close()
    perspective_popup = None

    # check viewer camera perspective value affects perspective popup slider
    # initial value
    viewer.camera.perspective = 10
    viewer.camera.zoom = 2
    viewer.camera.angles = (0, 0, 90)
    viewer_buttons.ndisplayButton.customContextMenuRequested.emit(QPoint())
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QtPopup):
            perspective_popup = widget
    assert perspective_popup
    assert viewer_buttons.perspective
    assert (
        viewer.camera.perspective == viewer_buttons.perspective.value() == 10
    )
    assert viewer_buttons.zoom
    assert viewer.camera.zoom == viewer_buttons.zoom.value() == 2
    assert viewer_buttons.rx
    assert viewer_buttons.ry
    assert viewer_buttons.rz
    assert (
        viewer.camera.angles
        == (
            viewer_buttons.rx.value(),
            viewer_buttons.ry.value(),
            viewer_buttons.rz.value(),
        )
        == (0, 0, 90)
    )


def test_toggle_ndisplay(mock_app_model, qt_viewer_buttons, qtbot):
    """Check `toggle_ndisplay` works via `mouseClick`."""
    viewer, viewer_buttons = qt_viewer_buttons
    assert viewer_buttons.ndisplayButton

    app = get_app_model()

    assert viewer.dims.ndisplay == 2
    with app.injection_store.register(
        providers=[
            (lambda: viewer, Viewer, 100),
        ]
    ):
        qtbot.mouseClick(viewer_buttons.ndisplayButton, Qt.LeftButton)
        assert viewer.dims.ndisplay == 3


def test_transpose_rotate_button(monkeypatch, qt_viewer_buttons, qtbot):
    """
    Click should trigger `transpose_axes`. Alt/Option-click should trigger `rotate_layers.`
    """
    _, viewer_buttons = qt_viewer_buttons
    assert viewer_buttons.transposeDimsButton

    action_manager_mock = Mock(trigger=Mock())

    # Monkeypatch the action_manager instance to prevent viewer error
    monkeypatch.setattr(
        'napari._qt.widgets.qt_viewer_buttons.action_manager',
        action_manager_mock,
    )
    modifiers = Qt.AltModifier
    qtbot.mouseClick(
        viewer_buttons.transposeDimsButton, Qt.LeftButton, modifiers
    )
    action_manager_mock.trigger.assert_called_with('napari:rotate_layers')

    trigger_mock = Mock()
    monkeypatch.setattr(
        'napari.utils.action_manager.ActionManager.trigger', trigger_mock
    )
    qtbot.mouseClick(viewer_buttons.transposeDimsButton, Qt.LeftButton)
    trigger_mock.assert_called_with('napari:transpose_axes')



import os
from unittest.mock import Mock

import numpy as np
import pytest
from vispy.app.canvas import MouseEvent

from napari._tests.utils import skip_on_win_ci
from napari.layers import Image
from napari.layers.base._base_constants import InteractionBoxHandle
from napari.layers.base._base_mouse_bindings import (
    highlight_box_handles,
    transform_with_box,
)


@skip_on_win_ci
def test_viewer_mouse_bindings(qtbot, make_napari_viewer):
    """Test adding mouse bindings to the viewer"""
    np.random.seed(0)
    viewer = make_napari_viewer()
    canvas = viewer.window._qt_viewer.canvas

    if os.getenv('CI'):
        viewer.show()

    mock_press = Mock()
    mock_drag = Mock()
    mock_release = Mock()
    mock_move = Mock()

    @viewer.mouse_drag_callbacks.append
    def drag_callback(v, event):
        assert viewer == v

        # on press
        mock_press.method()

        yield

        # on move
        while event.type == 'mouse_move':
            mock_drag.method()
            yield

        # on release
        mock_release.method()

    @viewer.mouse_move_callbacks.append
    def move_callback(v, event):
        assert viewer == v
        # on move
        mock_move.method()

    # Simulate press only
    canvas._scene_canvas.events.mouse_press(pos=(0, 0), modifiers=(), button=0)
    mock_press.method.assert_called_once()
    mock_press.reset_mock()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_not_called()
    mock_move.method.assert_not_called()

    # Simulate release only
    canvas._scene_canvas.events.mouse_release(
        pos=(0, 0), modifiers=(), button=0
    )
    mock_press.method.assert_not_called()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_called_once()
    mock_release.reset_mock()
    mock_move.method.assert_not_called()

    # Simulate move with no press
    canvas._scene_canvas.events.mouse_move(pos=(0, 0), modifiers=())
    mock_press.method.assert_not_called()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_not_called()
    mock_move.method.assert_called_once()
    mock_move.reset_mock()

    # Simulate press, drag, release
    canvas._scene_canvas.events.mouse_press(pos=(0, 0), modifiers=(), button=0)
    qtbot.wait(10)
    canvas._scene_canvas.events.mouse_move(
        pos=(0, 0),
        modifiers=(),
        button=0,
        press_event=MouseEvent(pos=(0, 0), type='mouse_press'),
    )
    qtbot.wait(10)
    canvas._scene_canvas.events.mouse_release(
        pos=(0, 0), modifiers=(), button=0
    )
    qtbot.wait(10)
    mock_press.method.assert_called_once()
    mock_drag.method.assert_called_once()
    mock_release.method.assert_called_once()
    mock_move.method.assert_not_called()


@skip_on_win_ci
def test_layer_mouse_bindings(qtbot, make_napari_viewer):
    """Test adding mouse bindings to a layer that is selected"""
    np.random.seed(0)
    viewer = make_napari_viewer()
    canvas = viewer.window._qt_viewer.canvas

    if os.getenv('CI'):
        viewer.show()

    layer = viewer.add_image(np.random.random((10, 20)))
    viewer.layers.selection.add(layer)

    mock_press = Mock()
    mock_drag = Mock()
    mock_release = Mock()
    mock_move = Mock()

    @layer.mouse_drag_callbacks.append
    def drag_callback(_layer, event):
        assert layer == _layer
        # on press
        mock_press.method()

        yield

        # on move
        while event.type == 'mouse_move':
            mock_drag.method()
            yield

        # on release
        mock_release.method()

    @layer.mouse_move_callbacks.append
    def move_callback(_layer, event):
        assert layer == _layer
        # on press
        mock_move.method()

    # Simulate press only
    canvas._scene_canvas.events.mouse_press(pos=(0, 0), modifiers=(), button=0)
    mock_press.method.assert_called_once()
    mock_press.reset_mock()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_not_called()
    mock_move.method.assert_not_called()

    # Simulate release only
    canvas._scene_canvas.events.mouse_release(
        pos=(0, 0), modifiers=(), button=0
    )
    mock_press.method.assert_not_called()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_called_once()
    mock_release.reset_mock()
    mock_move.method.assert_not_called()

    # Simulate move with no press
    canvas._scene_canvas.events.mouse_move(pos=(0, 0), modifiers=())
    mock_press.method.assert_not_called()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_not_called()
    mock_move.method.assert_called_once()
    mock_move.reset_mock()

    # Simulate press, drag, release
    canvas._scene_canvas.events.mouse_press(pos=(0, 0), modifiers=(), button=0)
    qtbot.wait(10)
    canvas._scene_canvas.events.mouse_move(
        pos=(0, 0),
        modifiers=(),
        button=0,
        press_event=MouseEvent(pos=(0, 0), type='mouse_press'),
    )
    qtbot.wait(10)
    canvas._scene_canvas.events.mouse_release(
        pos=(0, 0), modifiers=(), button=0
    )
    qtbot.wait(10)
    mock_press.method.assert_called_once()
    mock_drag.method.assert_called_once()
    mock_release.method.assert_called_once()
    mock_move.method.assert_not_called()


@skip_on_win_ci
def test_unselected_layer_mouse_bindings(qtbot, make_napari_viewer):
    """Test adding mouse bindings to a layer that is not selected"""
    np.random.seed(0)
    viewer = make_napari_viewer()
    canvas = viewer.window._qt_viewer.canvas

    if os.getenv('CI'):
        viewer.show()

    layer = viewer.add_image(np.random.random((10, 20)))
    viewer.layers.selection.remove(layer)

    mock_press = Mock()
    mock_drag = Mock()
    mock_release = Mock()
    mock_move = Mock()

    @layer.mouse_drag_callbacks.append
    def drag_callback(_layer, event):
        assert layer == _layer
        # on press
        mock_press.method()

        yield

        # on move
        while event.type == 'mouse_move':
            mock_drag.method()
            yield

        # on release
        mock_release.method()

    @layer.mouse_move_callbacks.append
    def move_callback(_layer, event):
        assert layer == _layer
        # on press
        mock_move.method()

    # Simulate press only
    canvas._scene_canvas.events.mouse_press(pos=(0, 0), modifiers=(), button=0)
    mock_press.method.assert_not_called()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_not_called()
    mock_move.method.assert_not_called()

    # Simulate release only
    canvas._scene_canvas.events.mouse_release(
        pos=(0, 0), modifiers=(), button=0
    )
    mock_press.method.assert_not_called()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_not_called()
    mock_move.method.assert_not_called()

    # Simulate move with no press
    canvas._scene_canvas.events.mouse_move(pos=(0, 0), modifiers=())
    mock_press.method.assert_not_called()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_not_called()
    mock_move.method.assert_not_called()

    # Simulate press, drag, release
    canvas._scene_canvas.events.mouse_press(pos=(0, 0), modifiers=(), button=0)
    qtbot.wait(10)
    canvas._scene_canvas.events.mouse_move(
        pos=(0, 0),
        modifiers=(),
        button=0,
        press_event=MouseEvent(pos=(0, 0), type='mouse_press'),
    )
    qtbot.wait(10)
    canvas._scene_canvas.events.mouse_release(
        pos=(0, 0), modifiers=(), button=0
    )
    qtbot.wait(10)
    mock_press.method.assert_not_called()
    mock_drag.method.assert_not_called()
    mock_release.method.assert_not_called()
    mock_move.method.assert_not_called()


@pytest.mark.parametrize(
    ('position', 'dims_displayed', 'nearby_handle'),
    [
        # Postion inside the transform box space so the inside value should be set as selected
        ([0, 3], [0, 1], InteractionBoxHandle.INSIDE),
        ([0, 3, 3], [1, 2], InteractionBoxHandle.INSIDE),
        # Postion outside the transform box space so no handle should be set as selected
        ([0, 11], [0, 1], None),
        ([0, 11, 11], [1, 2], None),
        # When 3 dimensions are being displayed no `highlight_box_handles` logic should be run
        ([0, 3, 3], [0, 1, 2], None),
    ],
)
def test_highlight_box_handles(position, dims_displayed, nearby_handle):
    layer = Image(np.empty((10, 10)))
    event = Mock(
        position=position, dims_displayed=dims_displayed, modifiers=[None]
    )
    highlight_box_handles(
        layer,
        event,
    )
    # mouse event should be detected over the expected handle
    assert layer._overlays['transform_box'].selected_handle == nearby_handle


def test_transform_box():
    layer = Image(np.empty((10, 10)))
    event = Mock(position=[0, 3], dims_displayed=[0, 1], modifiers=[None])
    next(transform_with_box(layer, event))
    # no interaction has been done so affine should be the same as the initial
    assert layer.affine == layer._initial_affine



import subprocess
from pathlib import Path
from typing import NamedTuple

from napari.utils import info


def test_citation_text():
    assert isinstance(info.citation_text, str)
    assert 'doi' in info.citation_text


def test_linux_os_name_file(monkeypatch, tmp_path):
    with open(tmp_path / 'os-release', 'w') as f_p:
        f_p.write('PRETTY_NAME="Test text"\n')

    monkeypatch.setattr(info, 'OS_RELEASE_PATH', str(tmp_path / 'os-release'))

    assert info._linux_sys_name() == 'Test text'

    with open(tmp_path / 'os-release', 'w') as f_p:
        f_p.write('NAME="Test2"\nVERSION="text"')

    assert info._linux_sys_name() == 'Test2 text'

    with open(tmp_path / 'os-release', 'w') as f_p:
        f_p.write('NAME="Test2"\nVERSION_ID="text2"')

    assert info._linux_sys_name() == 'Test2 text2'

    with open(tmp_path / 'os-release', 'w') as f_p:
        f_p.write('NAME="Test2"\nVERSION="text"\nVERSION_ID="text2"')

    assert info._linux_sys_name() == 'Test2 text'

    with open(tmp_path / 'os-release', 'w') as f_p:
        f_p.write(
            'PRETTY_NAME="Test text"\nNAME="Test2"\nVERSION="text"\nVERSION_ID="text2"'
        )

    assert info._linux_sys_name() == 'Test text'


class _CompletedProcessMock(NamedTuple):
    stdout: bytes


def _lsb_mock(*_args, **_kwargs):
    return _CompletedProcessMock(
        stdout=b'Description:	Ubuntu Test 20.04\nRelease:	20.04'
    )


def _lsb_mock2(*_args, **_kwargs):
    return _CompletedProcessMock(
        stdout=b'Description:	Ubuntu Test\nRelease:	20.05'
    )


def test_linux_os_name_lsb(monkeypatch, tmp_path):
    monkeypatch.setattr(info, 'OS_RELEASE_PATH', str(tmp_path / 'os-release'))
    monkeypatch.setattr(subprocess, 'run', _lsb_mock)
    assert info._linux_sys_name() == 'Ubuntu Test 20.04'
    monkeypatch.setattr(subprocess, 'run', _lsb_mock2)
    assert info._linux_sys_name() == 'Ubuntu Test 20.05'


def test_napari_from_conda_with_metadata(monkeypatch):
    """Test _napari_from_conda when the napari metadata file exists."""
    from napari import __version__

    def mock_glob(self, pattern):
        return [
            Path(f'/mock/conda-meta/napari-{__version__}-py2foo3bar.json'),
            Path('/mock/conda-meta/napari-svg-0.2.1-py3bar2foo.json'),
            Path('/mock/conda-meta/napari-console-0.1.3-py2foo3bar.json'),
        ]

    monkeypatch.setattr(Path, 'glob', mock_glob)
    assert info._napari_from_conda() is True


def test_napari_from_conda_without_metadata(monkeypatch):
    """Test _napari_from_conda when no napari metadata file exists."""

    def mock_glob(self, pattern):
        return [
            Path('/mock/conda-meta/napari-svg-0.2.1-py3bar2foo.json'),
            Path('/mock/conda-meta/napari-console-0.1.3-py2foo3bar.json'),
        ]

    monkeypatch.setattr(Path, 'glob', mock_glob)
    assert info._napari_from_conda() is False



from __future__ import annotations

from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtGui import QKeySequence, QPainter
from qtpy.QtWidgets import (
    QFormLayout,
    QLabel,
    QStackedWidget,
    QStyle,
    QStyleOption,
    QVBoxLayout,
    QWidget,
)

from napari import __version__
from napari.utils.action_manager import action_manager
from napari.utils.interactions import Shortcut
from napari.utils.translations import trans


class QtWelcomeLabel(QLabel):
    """Labels used for main message in welcome page."""


class QtShortcutLabel(QLabel):
    """Labels used for displaying shortcu information in welcome page."""


class QtVersionLabel(QLabel):
    """Label used for displaying version information."""


class QtWelcomeWidget(QWidget):
    """Welcome widget to display initial information and shortcuts to user."""

    sig_dropped = Signal('QEvent')

    def __init__(self, parent) -> None:
        super().__init__(parent)

        # Create colored icon using theme
        self._image = QLabel()
        self._image.setObjectName('logo_silhouette')
        self._image.setMinimumSize(300, 300)
        self._version_label = QtVersionLabel(f'napari {__version__}')
        self._label = QtWelcomeLabel(
            trans._(
                'Drag image(s) here to open\nor\nUse the menu shortcuts below:'
            )
        )

        # Widget setup
        self.setAutoFillBackground(True)
        self.setAcceptDrops(True)
        self._image.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._version_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        # Layout
        text_layout = QVBoxLayout()
        text_layout.setSpacing(10)
        text_layout.addWidget(self._version_label)
        text_layout.addWidget(self._label)

        # TODO: Use action manager for shortcut query and handling
        shortcut_layout = QFormLayout()
        sc = QKeySequence('Ctrl+N', QKeySequence.PortableText).toString(
            QKeySequence.NativeText
        )
        shortcut_layout.addRow(
            QtShortcutLabel(sc),
            QtShortcutLabel(trans._('New Image from Clipboard')),
        )
        sc = QKeySequence('Ctrl+O', QKeySequence.PortableText).toString(
            QKeySequence.NativeText
        )
        shortcut_layout.addRow(
            QtShortcutLabel(sc),
            QtShortcutLabel(trans._('Open image(s)')),
        )
        sc = QKeySequence('Ctrl+Shift+P', QKeySequence.PortableText).toString(
            QKeySequence.NativeText
        )
        shortcut_layout.addRow(
            QtShortcutLabel(sc),
            QtShortcutLabel(trans._('Show Command Palette')),
        )
        self._shortcut_label = QtShortcutLabel('')
        shortcut_layout.addRow(
            self._shortcut_label,
            QtShortcutLabel(trans._('Show all key bindings')),
        )
        shortcut_layout.setSpacing(0)

        layout = QVBoxLayout()
        layout.addStretch()
        layout.setSpacing(30)
        layout.addWidget(self._image)
        layout.addLayout(text_layout)
        layout.addLayout(shortcut_layout)
        layout.addStretch()

        self.setLayout(layout)
        self._show_shortcuts_updated()
        action_manager.events.shorcut_changed.connect(
            self._show_shortcuts_updated
        )

    def minimumSizeHint(self):
        """
        Overwrite minimum size to allow creating small viewer instance
        """
        return QSize(100, 100)

    def _show_shortcuts_updated(self):
        shortcut_list = list(
            action_manager._shortcuts['napari:show_shortcuts']
        )
        if not shortcut_list:
            return
        self._shortcut_label.setText(Shortcut(shortcut_list[0]).platform)

    def paintEvent(self, event):
        """Override Qt method.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        option = QStyleOption()
        option.initFrom(self)
        p = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, option, p, self)

    def _update_property(self, prop, value):
        """Update properties of widget to update style.

        Parameters
        ----------
        prop : str
            Property name to update.
        value : bool
            Property value to update.
        """
        self.setProperty(prop, value)
        self.style().unpolish(self)
        self.style().polish(self)

    def dragEnterEvent(self, event):
        """Override Qt method.

        Provide style updates on event.

        Parameters
        ----------
        event : qtpy.QtCore.QDragEnterEvent
            Event from the Qt context.
        """
        self._update_property('drag', True)
        if event.mimeData().hasUrls():
            viewer = self.parentWidget().nativeParentWidget()._qt_viewer
            viewer._set_drag_status()
            event.accept()
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        """Override Qt method.

        Provide style updates on event.

        Parameters
        ----------
        event : qtpy.QtCore.QDragLeaveEvent
            Event from the Qt context.
        """
        self._update_property('drag', False)

    def dropEvent(self, event):
        """Override Qt method.

        Provide style updates on event and emit the drop event.

        Parameters
        ----------
        event : qtpy.QtCore.QDropEvent
            Event from the Qt context.
        """
        self._update_property('drag', False)
        self.sig_dropped.emit(event)


class QtWidgetOverlay(QStackedWidget):
    """
    Stacked widget providing switching between the widget and a welcome page.
    """

    sig_dropped = Signal('QEvent')
    resized = Signal()
    leave = Signal()
    enter = Signal()

    def __init__(self, parent, widget) -> None:
        super().__init__(parent)

        self._overlay = QtWelcomeWidget(self)

        # Widget setup
        self.addWidget(widget)
        self.addWidget(self._overlay)
        self.setCurrentIndex(0)

        # Signals
        self._overlay.sig_dropped.connect(self.sig_dropped)

    def set_welcome_visible(self, visible=True):
        """Show welcome screen widget on stack."""
        self.setCurrentIndex(int(visible))

    def resizeEvent(self, event):
        """Emit our own event when canvas was resized."""
        self.resized.emit()
        return super().resizeEvent(event)

    def enterEvent(self, event):
        """Emit our own event when mouse enters the canvas."""
        self.enter.emit()
        super().enterEvent(event)

    def leaveEvent(self, event):
        """Emit our own event when mouse leaves the canvas."""
        self.leave.emit()
        super().leaveEvent(event)



import typing

from qtpy.QtCore import QModelIndex, QSize, Qt
from qtpy.QtGui import QImage

from napari import current_viewer
from napari._qt.containers.qt_list_model import QtListModel
from napari.layers import Layer
from napari.settings import get_settings
from napari.utils.translations import trans

ThumbnailRole = Qt.UserRole + 2
LoadedRole = Qt.UserRole + 3


class QtLayerListModel(QtListModel[Layer]):
    def data(self, index: QModelIndex, role: Qt.ItemDataRole):
        """Return data stored under ``role`` for the item at ``index``."""
        if not index.isValid():
            return None
        layer = self.getItem(index)
        viewer = current_viewer()
        layer_loaded = layer.loaded
        # Playback with async slicing causes flickering between the thumbnail
        # and loading animation in some cases due quick changes in the loaded
        # state, so report as unloaded in that case to avoid that.
        if get_settings().experimental.async_ and (viewer := current_viewer()):
            viewer_playing = viewer.window._qt_viewer.dims.is_playing
            layer_loaded = layer.loaded and not viewer_playing
        if role == Qt.ItemDataRole.DisplayRole:  # used for item text
            return layer.name
        if role == Qt.ItemDataRole.TextAlignmentRole:  # alignment of the text
            return Qt.AlignCenter
        if role == Qt.ItemDataRole.EditRole:
            # used to populate line edit when editing
            return layer.name
        if role == Qt.ItemDataRole.ToolTipRole:  # for tooltip
            layer_source_info = layer.get_source_str()
            if layer_loaded:
                return layer_source_info
            return trans._('{source} (loading)', source=layer_source_info)
        if (
            role == Qt.ItemDataRole.CheckStateRole
        ):  # the "checked" state of this item
            return (
                Qt.CheckState.Checked
                if layer.visible
                else Qt.CheckState.Unchecked
            )
        if role == Qt.ItemDataRole.SizeHintRole:  # determines size of item
            return QSize(200, 34)
        if role == ThumbnailRole:  # return the thumbnail
            thumbnail = layer.thumbnail
            return QImage(
                thumbnail,
                thumbnail.shape[1],
                thumbnail.shape[0],
                QImage.Format_RGBA8888,
            )
        if role == LoadedRole:
            return layer_loaded
        # normally you'd put the icon in DecorationRole, but we do that in the
        # # LayerDelegate which is aware of the theme.
        # if role == Qt.ItemDataRole.DecorationRole:  # icon to show
        #     pass
        return super().data(index, role)

    def setData(
        self,
        index: QModelIndex,
        value: typing.Any,
        role: int = Qt.ItemDataRole.EditRole,
    ) -> bool:
        if role == Qt.ItemDataRole.CheckStateRole:
            # The item model stores a Qt.CheckState enum value that can be
            # partially checked, but we only use the unchecked and checked
            # to correspond to the layer's visibility.
            # https://doc.qt.io/qt-5/qt.html#CheckState-enum
            self.getItem(index).visible = (
                Qt.CheckState(value) == Qt.CheckState.Checked
            )
        elif role == Qt.ItemDataRole.EditRole:
            self.getItem(index).name = value
            role = Qt.ItemDataRole.DisplayRole
        else:
            return super().setData(index, value, role=role)

        self.dataChanged.emit(index, index, [role])
        return True

    def all_loaded(self):
        """Return if all the layers are loaded."""
        return all(
            self.index(row, 0).data(LoadedRole)
            for row in range(self.rowCount())
        )

    def _process_event(self, event):
        # The model needs to emit `dataChanged` whenever data has changed
        # for a given index, so that views can update themselves.
        # Here we convert native events to the dataChanged signal.
        if not hasattr(event, 'index'):
            return
        role = {
            'thumbnail': ThumbnailRole,
            'visible': Qt.ItemDataRole.CheckStateRole,
            'name': Qt.ItemDataRole.DisplayRole,
            'loaded': LoadedRole,
        }.get(event.type)
        roles = [role] if role is not None else []
        row = self.index(event.index)
        self.dataChanged.emit(row, row, roles)



import pytest

from napari._app_model._app import get_app_model
from napari._app_model.actions._toggle_action import ViewerModelToggleAction
from napari._qt._qapp_model.qactions._toggle_action import (
    DockWidgetToggleAction,
)
from napari._qt.qt_main_window import Window
from napari.components import ViewerModel


@pytest.mark.usefixtures('mock_app_model')
def test_viewer_toggler():
    viewer = ViewerModel()
    action = ViewerModelToggleAction(
        id='some.command.id',
        title='Toggle Axis Visibility',
        viewer_attribute='axes',
        sub_attribute='visible',
    )
    app = get_app_model()
    app.register_action(action)

    # Injection required as there is no current viewer, use a high weight (100)
    # so this provider is used over `_provide_viewer`, which would raise an error
    with app.injection_store.register(
        providers=[
            (lambda: viewer, ViewerModel, 100),
        ]
    ):
        assert viewer.axes.visible is False
        app.commands.execute_command('some.command.id')
        assert viewer.axes.visible is True
        app.commands.execute_command('some.command.id')
        assert viewer.axes.visible is False


def test_dock_widget_toggler(make_napari_viewer):
    """Tests `DockWidgetToggleAction` toggling works."""
    viewer = make_napari_viewer(show=True)
    action = DockWidgetToggleAction(
        id='some.command.id',
        title='Toggle Dock Widget',
        dock_widget='dockConsole',
    )
    app = get_app_model()
    app.register_action(action)

    with app.injection_store.register(
        providers={Window: lambda: viewer.window}
    ):
        assert viewer.window._qt_viewer.dockConsole.isVisible() is False
        app.commands.execute_command('some.command.id')
        assert viewer.window._qt_viewer.dockConsole.isVisible() is True
        app.commands.execute_command('some.command.id')
        assert viewer.window._qt_viewer.dockConsole.isVisible() is False



"""A performant, dedicated thread to compute cursor status and signal updates to a viewer."""

from __future__ import annotations

import os
from threading import Event
from typing import TYPE_CHECKING
from weakref import ref

from qtpy.QtCore import QObject, QThread, Signal

from napari.utils.notifications import Notification, notification_manager

if TYPE_CHECKING:
    from napari.components import ViewerModel


class StatusChecker(QThread):
    """A dedicated thread for performant updating of the status bar.

    This class offloads the job of computing the cursor status into a separate thread,
    Qt Signals are used to update the main viewer with the status string.

    Prior to https://github.com/napari/napari/pull/7146, status bar updates
    happened on the main thread in the viewer model, which could be
    expensive in some layers and resulted in bad performance when some
    layers were selected.

    Because the thread runs a single infinite while loop, the updates are
    naturally throttled since they can only be sent at the rate which updates
    can be computed, but no faster.

    Attributes
    ----------
    _need_status_update : threading.Event
        An Event (fancy thread-safe bool-like to synchronize threads)
        for keeping track of when the status needs updating
        (because the cursor has moved).
    _terminate : bool
        If set to True, the status checker thread needs to be terminated.
        When the QtViewer is being closed, it sets this flag to terminate
        the status checker thread.
        After _terminate is set to True, no more status updates are sent.
        Default: False.
    viewer_ref : weakref.ref[napari.viewer.ViewerModel]
        A weak reference to the viewer which is providing status updates.
        We keep a weak reference to the viewer so the status checker thread
        will not prevent the viewer from being garbage collected.
        We proactively check the viewer to determine if a new status update
        needs to be computed and emitted.
    """

    # Create a Signal to establish a lightweight communication mechanism between the
    # viewer and the status checker thread for cursor events and related status
    status_and_tooltip_changed = Signal(object)

    def __init__(self, viewer: ViewerModel, parent: QObject | None = None):
        super().__init__(parent=parent)
        self.viewer_ref = ref(viewer)
        self._need_status_update = Event()
        self._need_status_update.clear()
        self._terminate = False
        self._allow_start = True
        self.setObjectName('StatusChecker')

    def trigger_status_update(self) -> None:
        """Trigger a status update computation.

        When the cursor moves, the viewer will call this to instruct
        the status checker to update the viewer with the present status.
        """
        self._need_status_update.set()

    def close_terminate(self) -> None:
        """Close the status checker thread.

        This method is called when the viewer is closed.
        """
        self._allow_start = False
        self.terminate()

    def terminate(self) -> None:
        """Terminate the status checker thread.

        For proper cleanup,it's important to set _terminate to True before
        calling _needs_status_update.set.
        """
        self._terminate = True
        self._need_status_update.set()

    def start(
        self, priority: QThread.Priority = QThread.Priority.InheritPriority
    ) -> None:
        """Start the status checker thread.

        Make sure to set the _terminate attribute to False prior to start.
        """
        if not self._allow_start:
            return
        self._terminate = False
        super().start(priority)

    def run(self) -> None:
        while not self._terminate:
            if self.viewer_ref() is None:
                # Stop thread when viewer is closed
                return
            if self._need_status_update.is_set():
                self._need_status_update.clear()
                self.calculate_status()
            else:
                self._need_status_update.wait()

    def calculate_status(self) -> None:
        """Calculate the status and emit the signal.

        If the viewer is not available, do nothing. Otherwise,
        emit the signal that the status has changed.
        """
        viewer = self.viewer_ref()
        if viewer is None:
            return

        try:
            # Calculate the status change from cursor's movement
            res = viewer._calc_status_from_cursor()
        except Exception as e:  # pragma: no cover # noqa: BLE001
            # Our codebase is not threadsafe. It is possible that an
            # ViewerModel or Layer state is changed while we are trying to
            # calculate the status, which may cause an exception.
            # All exceptions are caught and handled to keep updates
            # from crashing the thread. The exception is logged
            # and a notification is sent.
            notification_manager.dispatch(Notification.from_exception(e))
            return
        # Emit the signal with the updated status
        self.status_and_tooltip_changed.emit(res)


if os.environ.get('ASV') == 'true':
    # This is a hack to make sure that the StatusChecker thread is not
    # running when the benchmark is running. This is because the
    # StatusChecker thread may introduce some noise in the benchmark
    # results from waiting on its termination.
    StatusChecker.start = lambda self, priority=0: None  # type: ignore[assignment]



import subprocess
import sys
from typing import TYPE_CHECKING

import pytest
from napari_plugin_engine import napari_hook_implementation

if TYPE_CHECKING:
    from napari.plugins._plugin_manager import NapariPluginManager


def test_plugin_discovery_is_delayed():
    """Test that plugins are not getting discovered at napari import time."""
    cmd = [
        sys.executable,
        '-c',
        'import sys; from napari.plugins import plugin_manager; '
        'sys.exit(len(plugin_manager.plugins) > 2)',  # we have 2 'builtins'
    ]
    # will fail if plugin discovery happened at import
    proc = subprocess.run(cmd, capture_output=True)
    assert not proc.returncode, 'Plugins were discovered at import time!'


def test_plugin_events(napari_plugin_manager):
    """Test event emission by plugin manager."""
    tnpm: NapariPluginManager = napari_plugin_manager

    register_events = []
    unregister_events = []
    enable_events = []
    disable_events = []

    tnpm.events.registered.connect(lambda e: register_events.append(e))
    tnpm.events.unregistered.connect(lambda e: unregister_events.append(e))
    tnpm.events.enabled.connect(lambda e: enable_events.append(e))
    tnpm.events.disabled.connect(lambda e: disable_events.append(e))

    class Plugin:
        pass

    tnpm.register(Plugin, name='Plugin')
    assert 'Plugin' in tnpm.plugins
    assert len(register_events) == 1
    assert register_events[0].value == 'Plugin'
    assert not enable_events
    assert not disable_events

    tnpm.unregister(Plugin)
    assert len(unregister_events) == 1
    assert unregister_events[0].value == 'Plugin'

    tnpm.set_blocked('Plugin')
    assert len(disable_events) == 1
    assert disable_events[0].value == 'Plugin'
    assert not enable_events
    assert 'Plugin' not in tnpm.plugins
    # blocked from registering
    assert tnpm.is_blocked('Plugin')
    tnpm.register(Plugin, name='Plugin')
    assert 'Plugin' not in tnpm.plugins
    assert len(register_events) == 1

    tnpm.set_blocked('Plugin', False)
    assert not tnpm.is_blocked('Plugin')
    assert len(enable_events) == 1
    assert enable_events[0].value == 'Plugin'
    # note: it doesn't immediately re-register it
    assert 'Plugin' not in tnpm.plugins
    # but we can now re-register it
    tnpm.register(Plugin, name='Plugin')
    assert len(register_events) == 2


def test_plugin_extension_assignment(napari_plugin_manager):
    class Plugin:
        @napari_hook_implementation
        def napari_get_reader(path):
            if path.endswith('.png'):
                return lambda x: None
            return None

        @napari_hook_implementation
        def napari_get_writer(path, *args):
            if path.endswith('.png'):
                return lambda x: None
            return None

    tnpm: NapariPluginManager = napari_plugin_manager
    tnpm.register(Plugin, name='test_plugin')

    assert tnpm.get_reader_for_extension('.png') is None
    tnpm.assign_reader_to_extensions('test_plugin', '.png')
    assert '.png' in tnpm._extension2reader
    assert tnpm.get_reader_for_extension('.png') == 'test_plugin'

    with pytest.warns(UserWarning, match='did not return a reader function'):
        # reader may not recognize extension
        tnpm.assign_reader_to_extensions('test_plugin', '.pndfdg')

    with pytest.raises(ValueError, match='is not a valid reader'):
        tnpm.assign_reader_to_extensions('test_pldfdfugin', '.png')



from unittest.mock import MagicMock

import numpy as np
import pandas as pd
import pytest
from pandas.core.generic import pandas_dtype
from qtpy.QtCore import QItemSelection, QItemSelectionModel, Qt
from qtpy.QtGui import QGuiApplication
from qtpy.QtWidgets import (
    QAbstractItemDelegate,
    QComboBox,
    QFileDialog,
    QLineEdit,
)

from napari.components import ViewerModel
from napari_builtins._qt.features_table import FeaturesTable, PandasModel


@pytest.mark.usefixtures('qapp')
def test_pandas_model():
    df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [True, False, True, False]})
    view = PandasModel(df)
    assert view.rowCount() == 4
    assert view.columnCount() == 3  # 0 is index
    assert view.data(view.index(0, 1)) == '1'
    assert view.headerData(1, Qt.Orientation.Horizontal) == 'a'
    assert view.headerData(2, Qt.Orientation.Horizontal) == 'b'
    assert view.headerData(1, Qt.Orientation.Vertical) == 1


@pytest.mark.usefixtures('qapp')
def test_pandas_model_flags():
    df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [True, False, True, False]})
    view = PandasModel(df)
    assert view.flags(view.index(5, 5)) == Qt.ItemFlag.ItemIsEnabled
    assert view.flags(view.index(5, 5)) == Qt.ItemFlag.ItemIsEnabled

    view.editable = True
    assert view.flags(view.index(0, 1)) & Qt.ItemFlag.ItemIsEditable
    assert view.flags(view.index(0, 2)) & Qt.ItemFlag.ItemIsUserCheckable


def test_pandas_model_set_data(qtbot):
    df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [True, False, True, False]})
    view = PandasModel(df)
    assert not view.setData(view.index(5, 5), 1)
    assert not view.setData(view.index(0, 0), 1)  # cannot edit index
    assert view.setData(view.index(0, 1), 5)
    assert df.loc[0, 'a'] == 5
    assert not view.setData(view.index(0, 1), 7, Qt.ItemDataRole.UserRole)
    assert not view.setData(
        view.index(0, 1), 'aaa'
    )  # cannot set string to int column

    with qtbot.waitSignal(view.dataChanged):
        assert view.setData(view.index(0, 2), False)
    assert not df.loc[0, 'b']

    with qtbot.waitSignal(view.dataChanged):
        assert view.setData(
            view.index(0, 2),
            Qt.CheckState.Checked,
            Qt.ItemDataRole.CheckStateRole,
        )
    assert df.loc[0, 'b']


def test_pandas_model_set_data_categorical(qtbot):
    df = pd.DataFrame(
        {
            'a': pd.Series(['a', 'b', 'a', 'b'], dtype='category'),
            'b': [1, 2, 3, 4],
        }
    )
    view = PandasModel(df)

    with qtbot.waitSignal(view.dataChanged):
        assert view.setData(view.index(0, 1), 'b')
    assert df.loc[0, 'a'] == 'b'

    with qtbot.waitSignal(view.dataChanged):
        assert view.setData(view.index(0, 1), 'a')
    assert df.loc[0, 'a'] == 'a'

    assert not view.setData(view.index(0, 1), 'c')  # not in categories


def test_features_table(qtbot):
    v = ViewerModel()
    w = FeaturesTable(v)
    qtbot.add_widget(w)
    proxy = w.table.model()

    assert proxy.columnCount() == 1  # 0 is index
    assert proxy.rowCount() == 0

    original_a = (2, 0, 1)

    layer = v.add_points(np.zeros((3, 2)), features={'a': original_a})
    assert proxy.columnCount() == 2
    assert proxy.rowCount() == 3

    layer.features['b'] = [True, False, True]
    layer.events.features()

    assert proxy.columnCount() == 3
    assert proxy.rowCount() == 3

    # sorting should sort the proxy model but not the data
    w.table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    for i in range(3):
        assert proxy.data(proxy.index(i, 1), Qt.ItemDataRole.EditRole) == str(
            i
        )
        assert layer.features['a'][i] == original_a[i]

    # sorting bools should work
    w.table.sortByColumn(2, Qt.SortOrder.AscendingOrder)
    for i in range(3):
        assert (
            proxy.data(
                proxy.index(
                    i,
                    2,
                ),
                Qt.ItemDataRole.EditRole,
            )
            == sorted(layer.features['b'])[i]
        )

    # test selection (with sorted rows)
    layer.selected_data = {1, 2}

    selected_rows = {
        proxy.mapToSource(raw_index).row()
        for raw_index in w.table.selectionModel().selectedIndexes()
    }
    assert selected_rows == layer.selected_data

    # click on top left cell
    idx = proxy.mapFromSource(proxy.sourceModel().index(0, 0))

    # select whole rows
    first_cell = proxy.index(idx.row(), 0)
    last_cell = proxy.index(idx.row(), proxy.columnCount() - 1)
    selection = QItemSelection()
    selection.select(first_cell, last_cell)

    w.table.selectionModel().select(
        selection,
        QItemSelectionModel.SelectionFlag.ClearAndSelect
        | QItemSelectionModel.SelectionFlag.Rows,
    )

    assert layer.selected_data == {0}


def test_features_table_selection_labels(qtbot):
    v = ViewerModel()
    w = FeaturesTable(v)
    qtbot.add_widget(w)

    original_a = (2, 0, 1)

    layer = v.add_labels(
        np.zeros((10, 10), dtype=np.uint8), features={'a': original_a}
    )
    assert layer.selected_label == 1

    w.table.selectRow(2)
    assert layer.selected_label == 2


def test_features_table_edit(qtbot):
    v = ViewerModel()
    w = FeaturesTable(v)
    qtbot.add_widget(w)
    proxy = w.table.model()

    original_a = ['x', 'y']

    layer = v.add_points(np.zeros((2, 2)), features={'a': original_a})

    idx = proxy.index(0, 1)
    w.table.edit(idx)
    assert not w.table.isPersistentEditorOpen(idx)

    w.toggle.click()
    assert proxy.sourceModel().editable
    w.table.edit(idx)
    assert w.table.isPersistentEditorOpen(idx)

    editor = w.table.findChild(QLineEdit)
    qtbot.keyClicks(editor, 'hello')
    assert editor.text() == 'hello'
    w.table.commitData(editor)
    assert layer.features.loc[0, 'a'] == 'hello'


def test_features_table_save_csv(qtbot, tmp_path, monkeypatch):
    v = ViewerModel()
    w = FeaturesTable(v)
    qtbot.add_widget(w)

    df = pd.DataFrame({'a': [1, 2]})
    v.add_points(np.zeros((2, 2)), features=df)

    path = tmp_path / 'test.csv'
    monkeypatch.setattr(
        QFileDialog, 'getSaveFileName', MagicMock(return_value=(path, None))
    )

    w.save.click()

    pd.testing.assert_frame_equal(pd.read_csv(path, index_col=0), df)


def test_features_table_copy_paste(qtbot, qapp):
    v = ViewerModel()
    w = FeaturesTable(v)
    qtbot.add_widget(w)
    proxy = w.table.model()

    df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})

    layer = v.add_points(np.zeros((3, 2)), features=df.copy())

    first_cell = proxy.index(1, 1)
    last_cell = proxy.index(1, 2)
    selection = QItemSelection()
    selection.select(first_cell, last_cell)

    w.table.selectionModel().select(
        selection,
        QItemSelectionModel.SelectionFlag.ClearAndSelect,
    )
    qapp.clipboard().setText('')
    qtbot.keyClick(w.table, 'c')
    assert qapp.clipboard().text().strip() == ''
    qtbot.keyClick(w.table, 'c', Qt.KeyboardModifier.ControlModifier)

    # stip cause windows and linux otherwise differ
    assert qapp.clipboard().text().strip() == '2\t5'

    first_cell = proxy.index(2, 1)
    last_cell = proxy.index(2, 2)
    selection = QItemSelection()
    selection.select(first_cell, last_cell)

    w.table.selectionModel().select(
        selection,
        QItemSelectionModel.SelectionFlag.ClearAndSelect,
    )

    w.toggle.click()
    QGuiApplication.clipboard().setText('3\t8\t7')
    # we test here that presence of additional columns does not
    # cause issues when pasting and we just discard them
    qtbot.keyClick(w.table, 'v', Qt.KeyboardModifier.ControlModifier)

    np.testing.assert_array_equal(layer.features.iloc[2], [3, 8])


@pytest.mark.parametrize(
    ('dtype', 'val', 'rendered_val', 'editor_class', 'new_val'),
    [
        (int, 2, '2', QLineEdit, 3),
        (float, 123.45678, '123.457', QLineEdit, 1e10),
        (
            'datetime64[ns]',
            '22-07-2025',
            '2025-07-22',
            QLineEdit,
            '2025-03-14',
        ),
        (bool, False, '', None, None),  # bool uses checkboxes instead
        (bool, True, '', None, None),  # bool uses checkboxes instead
        (pd.CategoricalDtype(['x', 'y']), 'x', 'x', QComboBox, 'y'),
    ],
)
def test_features_tables_dtypes(
    dtype, val, rendered_val, editor_class, new_val, qtbot
):
    v = ViewerModel()
    w = FeaturesTable(v)
    qtbot.add_widget(w)
    proxy = w.table.model()

    df = pd.DataFrame({'a': pd.Series([val], dtype=dtype)})

    layer = v.add_points(np.zeros((1, 2)), features=df)
    idx = proxy.index(
        0,
        1,
    )
    assert layer.features['a'].dtype == pandas_dtype(dtype)
    assert (
        proxy.data(
            idx,
            Qt.ItemDataRole.DisplayRole,
        )
        == rendered_val
    )

    w.toggle.click()
    w.table.edit(idx)

    if editor_class is None:
        # bools use checkboxes and not an editor, skip
        return

    editor = w.table.findChild(editor_class)

    if issubclass(editor_class, QLineEdit):
        qtbot.keyClicks(editor, str(new_val))
        w.table.commitData(editor)
    elif editor_class == QComboBox:
        qtbot.keyClick(editor, Qt.Key.Key_Down)
        w.table.commitData(editor)
    qtbot.wait(5)  # wait on singleShot execution
    w.table.closeEditor(editor, QAbstractItemDelegate.EndEditHint.NoHint)

    assert (
        layer.features.loc[0, 'a']
        == pd.Series(new_val, dtype=pandas_dtype(dtype))[0]
    )


def test_features_table_change_active_layer(qtbot):
    v = ViewerModel()
    w = FeaturesTable(v)
    qtbot.add_widget(w)

    df = pd.DataFrame({'a': [1, 2, 3]})
    layer1 = v.add_points(np.zeros((3, 2)), features=df.copy())
    layer2 = v.add_labels(
        np.zeros((10, 10), dtype=np.uint8), features=df.copy()
    )
    layer3 = v.add_image(np.empty((10, 10), dtype=np.uint8))

    v.layers.selection.active = layer1
    assert len(layer1.events.features.callbacks) == 2
    assert len(layer1.selected_data.events.all) == 1
    assert len(layer2.events.features.callbacks) == 1
    assert len(layer2.events.selected_label.callbacks) == 1
    assert w.info.text() == f'Features of "{layer1.name}"'

    v.layers.selection.active = layer2
    assert len(layer1.events.features.callbacks) == 1
    assert len(layer1.selected_data.events.all) == 0
    assert len(layer2.events.features.callbacks) == 2
    assert len(layer2.events.selected_label.callbacks) == 2
    assert w.info.text() == f'Features of "{layer2.name}"'

    v.layers.selection.active = layer3
    assert len(layer1.events.features.callbacks) == 1
    assert len(layer1.selected_data.events.all) == 0
    assert len(layer2.events.features.callbacks) == 1
    assert len(layer2.events.selected_label.callbacks) == 1
    assert 'has no features table' in w.info.text()

    v.layers.selection.active = None
    assert len(layer1.events.features.callbacks) == 1
    assert len(layer2.events.features.callbacks) == 1
    assert 'No layer selected.' in w.info.text()



"""Qt 'Window' menu Actions."""

from app_model.types import Action

from napari._app_model.constants import MenuGroup, MenuId
from napari._qt._qapp_model.qactions._toggle_action import (
    DockWidgetToggleAction,
)
from napari.utils.translations import trans

Q_WINDOW_ACTIONS: list[Action] = []

toggle_action_details = [
    (
        'napari:window:window:toggle_window_console',
        trans._('Console'),
        'dockConsole',
        trans._('Toggle console panel'),
    ),
    (
        'napari:window:window:toggle_layer_controls',
        trans._('Layer Controls'),
        'dockLayerControls',
        trans._('Toggle layer controls panel'),
    ),
    (
        'napari:window:window:toggle_layer_list',
        trans._('Layer List'),
        'dockLayerList',
        trans._('Toggle layer list panel'),
    ),
]
for cmd_id, cmd_title, dock_widget, status_tip in toggle_action_details:
    Q_WINDOW_ACTIONS.append(
        DockWidgetToggleAction(
            id=cmd_id,
            title=cmd_title,
            dock_widget=dock_widget,
            menus=[
                {
                    'id': MenuId.MENUBAR_WINDOW,
                    'group': MenuGroup.NAVIGATION,
                }
            ],
            status_tip=status_tip,
        )
    )



import warnings
from enum import Enum, EnumMeta
from functools import partial, wraps
from typing import TYPE_CHECKING

from qtpy.QtCore import QEvent, Qt
from qtpy.QtWidgets import (
    QApplication,
    QDoubleSpinBox,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QWidget,
)
from superqt import QEnumComboBox, QLabeledDoubleSlider

from napari._qt.dialogs.qt_modal import QtPopup
from napari._qt.widgets.qt_dims_sorter import QtDimsSorter
from napari._qt.widgets.qt_spinbox import QtSpinBox
from napari._qt.widgets.qt_tooltip import QtToolTipLabel
from napari.utils.action_manager import action_manager
from napari.utils.camera_orientations import (
    DepthAxisOrientation,
    DepthAxisOrientationStr,
    HorizontalAxisOrientation,
    HorizontalAxisOrientationStr,
    VerticalAxisOrientation,
    VerticalAxisOrientationStr,
)
from napari.utils.misc import in_ipython, in_jupyter, in_python_repl
from napari.utils.translations import trans

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

    from napari.viewer import ViewerModel


def add_new_points(viewer):
    viewer.add_points(
        ndim=max(viewer.dims.ndim, 2),
        scale=viewer.layers.extent.step,
    )


def add_new_shapes(viewer):
    viewer.add_shapes(
        ndim=max(viewer.dims.ndim, 2),
        scale=viewer.layers.extent.step,
    )


class QtLayerButtons(QFrame):
    """Button controls for napari layers.

    Parameters
    ----------
    viewer : napari.components.ViewerModel
        Napari viewer containing the rendered scene, layers, and controls.

    Attributes
    ----------
    deleteButton : QtDeleteButton
        Button to delete selected layers.
    newLabelsButton : QtViewerPushButton
        Button to add new Label layer.
    newPointsButton : QtViewerPushButton
        Button to add new Points layer.
    newShapesButton : QtViewerPushButton
        Button to add new Shapes layer.
    viewer : napari.components.ViewerModel
        Napari viewer containing the rendered scene, layers, and controls.
    """

    def __init__(self, viewer: 'ViewerModel') -> None:
        super().__init__()

        self.viewer = viewer

        self.deleteButton = QtViewerPushButton(
            'delete_button', action='napari:delete_selected_layers'
        )

        self.newPointsButton = QtViewerPushButton(
            'new_points',
            trans._('New points layer'),
            partial(add_new_points, self.viewer),
        )

        self.newShapesButton = QtViewerPushButton(
            'new_shapes',
            trans._('New shapes layer'),
            partial(add_new_shapes, self.viewer),
        )
        self.newLabelsButton = QtViewerPushButton(
            'new_labels',
            trans._('New labels layer'),
            self.viewer._new_labels,
        )

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.newPointsButton)
        layout.addWidget(self.newShapesButton)
        layout.addWidget(self.newLabelsButton)
        layout.addStretch(0)
        layout.addWidget(self.deleteButton)
        self.setLayout(layout)


def labeled_double_slider(
    *,
    parent: QtPopup,
    value: float,
    value_range: tuple[float, float],
    decimals: int = 0,
    callback: 'Callable',
) -> QLabeledDoubleSlider:
    """Create a labeled double slider widget."""
    slider = QLabeledDoubleSlider(parent)
    slider.setRange(*value_range)
    slider.setValue(value)
    slider.setDecimals(decimals)
    slider.valueChanged.connect(callback)
    return slider


def enum_combobox(
    *,
    parent: QtPopup,
    enum_class: EnumMeta,
    current_enum: Enum,
    callback: 'Callable[[],Any] | Callable[[Enum],Any]',
) -> QEnumComboBox:
    """Create an enum combobox widget."""
    combo = QEnumComboBox(parent, enum_class=enum_class)
    combo.setCurrentEnum(current_enum)
    combo.currentEnumChanged.connect(callback)
    return combo


def help_tooltip(
    *,
    parent: QtPopup,
    text: str,
    object_name: str = 'help_label',
) -> QtToolTipLabel:
    """Create a help tooltip widget."""
    help_symbol = QtToolTipLabel(parent)
    help_symbol.setObjectName(object_name)
    help_symbol.setToolTip(text)
    return help_symbol


class QtViewerButtons(QFrame):
    """Button controls for the napari viewer.

    Parameters
    ----------
    viewer : napari.components.ViewerModel
        Napari viewer containing the rendered scene, layers, and controls.

    Attributes
    ----------
    consoleButton : QtViewerPushButton
        Button to open iPython console within napari.
    rollDimsButton : QtViewerPushButton
        Button to roll orientation of spatial dimensions in the napari viewer.
    transposeDimsButton : QtViewerPushButton
        Button to transpose dimensions in the napari viewer.
    resetViewButton : QtViewerPushButton
        Button resetting the view of the rendered scene.
    gridViewButton : QtViewerPushButton
        Button to toggle grid view mode of layers on and off.
    ndisplayButton : QtViewerPushButton
        Button to toggle number of displayed dimensions.
    viewer : napari.components.ViewerModel
        Napari viewer containing the rendered scene, layers, and controls.
    """

    def __init__(self, viewer: 'ViewerModel') -> None:
        super().__init__()

        self.viewer = viewer

        self.consoleButton = QtViewerPushButton(
            'console', action='napari:toggle_console_visibility'
        )
        self.consoleButton.setProperty('expanded', False)
        if in_ipython() or in_jupyter() or in_python_repl():
            self.consoleButton.setEnabled(False)

        rdb = QtViewerPushButton('roll', action='napari:roll_axes')
        self.rollDimsButton = rdb
        rdb.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        rdb.customContextMenuRequested.connect(self._open_roll_popup)

        self.transposeDimsButton = QtViewerPushButton(
            'transpose',
            action='napari:transpose_axes',
            extra_tooltip_text=trans._(
                '\nAlt/option-click to rotate visible axes'
            ),
        )
        self.transposeDimsButton.installEventFilter(self)

        self.resetViewButton = QtViewerPushButton(
            'home', action='napari:reset_view'
        )
        gvb = QtViewerPushButton(
            'grid_view_button', action='napari:toggle_grid'
        )
        self.gridViewButton = gvb
        gvb.setCheckable(True)
        gvb.setChecked(viewer.grid.enabled)
        gvb.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        gvb.customContextMenuRequested.connect(self._open_grid_popup)

        @self.viewer.grid.events.enabled.connect
        def _set_grid_mode_checkstate(event):
            gvb.setChecked(event.value)

        ndb = QtViewerPushButton(
            'ndisplay_button', action='napari:toggle_ndisplay'
        )
        self.ndisplayButton = ndb
        ndb.setCheckable(True)
        ndb.setChecked(self.viewer.dims.ndisplay == 3)
        ndb.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        ndb.customContextMenuRequested.connect(self.open_ndisplay_camera_popup)

        @self.viewer.dims.events.ndisplay.connect
        def _set_ndisplay_mode_checkstate(event):
            ndb.setChecked(event.value == 3)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.consoleButton)
        layout.addWidget(self.ndisplayButton)
        layout.addWidget(self.rollDimsButton)
        layout.addWidget(self.transposeDimsButton)
        layout.addWidget(self.gridViewButton)
        layout.addWidget(self.resetViewButton)
        layout.addStretch(0)
        self.setLayout(layout)

    def eventFilter(self, qobject, event):
        """Have Alt/Option key rotate layers with the transpose button."""
        modifiers = QApplication.keyboardModifiers()
        if (
            modifiers == Qt.AltModifier
            and qobject == self.transposeDimsButton
            and event.type() == QEvent.MouseButtonPress
        ):
            action_manager.trigger('napari:rotate_layers')
            return True
        return False

    def _position_popup_inside_viewer(
        self, popup: QtPopup, button: QPushButton
    ) -> None:
        """Position the popup so that it remains with the viewer.

        Use the location of the button as a reference and place the popup
        above it. move_to will adjust the position if the popup is too close
        to the edge of the screen.
        """
        button_rect = button.rect()
        button_pos = button.mapToGlobal(button_rect.topLeft())
        popup.move_to(
            (
                button_pos.x(),
                button_pos.y() - popup.sizeHint().height() - 5,
                popup.sizeHint().width(),
                popup.sizeHint().height(),
            )
        )
        popup.show()

    def _add_3d_camera_controls(
        self,
        popup: QtPopup,
        grid_layout: QGridLayout,
    ) -> None:
        """Add 3D camera controls to the popup."""
        self.perspective = labeled_double_slider(
            parent=popup,
            value=self.viewer.camera.perspective,
            value_range=(0, 90),
            callback=self._update_perspective,
        )

        perspective_help_symbol = help_tooltip(
            parent=popup,
            text='Controls perspective projection strength. 0 is orthographic, larger values increase perspective effect.',
        )

        self.rx = labeled_double_slider(
            parent=popup,
            value=self.viewer.camera.angles[0],
            value_range=(-180, 180),
            callback=partial(self._update_camera_angles, 0),
        )

        self.ry = labeled_double_slider(
            parent=popup,
            value=self.viewer.camera.angles[1],
            value_range=(-89, 89),
            callback=partial(self._update_camera_angles, 1),
        )

        self.rz = labeled_double_slider(
            parent=popup,
            value=self.viewer.camera.angles[2],
            value_range=(-180, 180),
            callback=partial(self._update_camera_angles, 2),
        )

        angle_help_symbol = help_tooltip(
            parent=popup,
            text='Controls the rotation angles around each axis in degrees.',
        )

        grid_layout.addWidget(QLabel(trans._('Perspective:')), 2, 0)
        grid_layout.addWidget(self.perspective, 2, 1)
        grid_layout.addWidget(perspective_help_symbol, 2, 2)

        grid_layout.addWidget(QLabel(trans._('Angles    X:')), 3, 0)
        grid_layout.addWidget(self.rx, 3, 1)
        grid_layout.addWidget(angle_help_symbol, 3, 2)

        grid_layout.addWidget(QLabel(trans._('             Y:')), 4, 0)
        grid_layout.addWidget(self.ry, 4, 1)

        grid_layout.addWidget(QLabel(trans._('             Z:')), 5, 0)
        grid_layout.addWidget(self.rz, 5, 1)

    def _add_shared_camera_controls(
        self,
        popup: QtPopup,
        grid_layout: QGridLayout,
    ) -> None:
        """Add shared camera controls to the popup."""
        self.zoom = labeled_double_slider(
            parent=popup,
            value=self.viewer.camera.zoom,
            value_range=(0.01, 100),
            decimals=2,
            callback=self._update_zoom,
        )

        zoom_help_symbol = help_tooltip(
            parent=popup,
            text='Controls zoom level of the camera. Larger values zoom in, smaller values zoom out.',
        )

        grid_layout.addWidget(QLabel(trans._('Zoom:')), 1, 0)
        grid_layout.addWidget(self.zoom, 1, 1)
        grid_layout.addWidget(zoom_help_symbol, 1, 2)

    def _add_orientation_controls(
        self,
        popup: QtPopup,
        grid_layout: QGridLayout,
    ) -> None:
        """Add orientation controls to the popup."""
        orientation_widget = QWidget(popup)
        orientation_layout = QHBoxLayout()
        orientation_layout.setContentsMargins(0, 0, 0, 0)

        self.vertical_combo = enum_combobox(
            parent=popup,
            enum_class=VerticalAxisOrientation,
            current_enum=self.viewer.camera.orientation[1],
            callback=partial(
                self._update_orientation, VerticalAxisOrientation
            ),
        )

        self.horizontal_combo = enum_combobox(
            parent=popup,
            enum_class=HorizontalAxisOrientation,
            current_enum=self.viewer.camera.orientation[2],
            callback=partial(
                self._update_orientation, HorizontalAxisOrientation
            ),
        )

        if self.viewer.dims.ndisplay == 2:
            orientation_layout.addWidget(self.vertical_combo)
            orientation_layout.addWidget(self.horizontal_combo)
            self.orientation_help_symbol = help_tooltip(
                parent=popup,
                text='Controls the orientation of the vertical and horizontal camera axes.',
            )

        else:
            self.depth_combo = enum_combobox(
                parent=popup,
                enum_class=DepthAxisOrientation,
                current_enum=self.viewer.camera.orientation[0],
                callback=partial(
                    self._update_orientation, DepthAxisOrientation
                ),
            )

            orientation_layout.addWidget(self.depth_combo)
            orientation_layout.addWidget(self.vertical_combo)
            orientation_layout.addWidget(self.horizontal_combo)

            self.orientation_help_symbol = help_tooltip(
                parent=popup,
                text='',  # updated dynamically
            )
            self._update_handedness_help_symbol()
            self.viewer.camera.events.orientation.connect(
                self._update_handedness_help_symbol
            )

        orientation_widget.setLayout(orientation_layout)

        grid_layout.addWidget(QLabel(trans._('Orientation:')), 0, 0)
        grid_layout.addWidget(orientation_widget, 0, 1)
        grid_layout.addWidget(self.orientation_help_symbol, 0, 2)

    def _update_handedness_help_symbol(self, event=None) -> None:
        """Update the handedness symbol based on the camera orientation."""
        handedness = self.viewer.camera.handedness
        tooltip_text = (
            'Controls the orientation of the depth, vertical, and horizontal camera axes.\n'
            'Default is right-handed (towards, down, right).\n'
            'Default prior to 0.6.0 was left-handed (away, down, right).\n'
            f'Currently orientation is {handedness.value}-handed.'
        )
        self.orientation_help_symbol.setToolTip(tooltip_text)
        self.orientation_help_symbol.setObjectName(
            f'{handedness.value}hand_label'
        )
        self.orientation_help_symbol.style().polish(
            self.orientation_help_symbol
        )

    def open_ndisplay_camera_popup(self) -> None:
        """Show controls for camera settings based on ndisplay mode."""
        popup = QtPopup(self)
        grid_layout = QGridLayout()

        # Add orientation controls
        self._add_orientation_controls(popup, grid_layout)

        # Add shared camera controls
        self._add_shared_camera_controls(popup, grid_layout)

        # Add 3D camera controls if in 3D mode
        if self.viewer.dims.ndisplay == 3:
            self._add_3d_camera_controls(popup, grid_layout)

        popup.frame.setLayout(grid_layout)

        # Reposition popup, must be done after all widgets are added
        self._position_popup_inside_viewer(popup, self.ndisplayButton)

    def _update_orientation(
        self,
        orientation_type: type[
            DepthAxisOrientation
            | VerticalAxisOrientation
            | HorizontalAxisOrientation
        ],
        orientation_value: (
            DepthAxisOrientationStr
            | VerticalAxisOrientationStr
            | HorizontalAxisOrientationStr
        ),
    ) -> None:
        """Update the orientation of the camera.

        Parameters
        ----------
        orientation_type : type[DepthAxisOrientation | VerticalAxisOrientation | HorizontalAxisOrientation]
            The orientation type (which implies the position/axis) to update.
        value : DepthAxisOrientationStr | VerticalAxisOrientationStr | HorizontalAxisOrientationStr
            New orientation value for the updated axis.
        """
        axes = (
            DepthAxisOrientation,
            VerticalAxisOrientation,
            HorizontalAxisOrientation,
        )
        axis_to_update = axes.index(orientation_type)
        new_orientation = list(self.viewer.camera.orientation)
        new_orientation[axis_to_update] = orientation_type(orientation_value)
        self.viewer.camera.orientation = tuple(new_orientation)

    def _update_camera_angles(self, idx: int, value: float) -> None:
        """Update the camera angles.

        Parameters
        ----------
        idx : int
            Index of the angle to update. In the order of (rx, ry, rz).
        value : float
            New angle value.
        """

        angles = list(self.viewer.camera.angles)
        angles[idx] = value
        self.viewer.camera.angles = tuple(angles)

    def _update_zoom(self, value: float) -> None:
        """Update the camera zoom.

        Parameters
        ----------
        value : float
            New camera.zoom value.
        """

        self.viewer.camera.zoom = value

    def _update_perspective(self, value: float) -> None:
        """Update the camera perspective.

        Parameters
        ----------
        value : float
            New camera.perspective value.
        """

        self.viewer.camera.perspective = value

    def _open_roll_popup(self):
        """Open a grid popup to manually order the dimensions"""
        pop = QtPopup(self)

        # dims sorter widget
        dim_sorter = QtDimsSorter(self.viewer.dims, pop)
        dim_sorter.setObjectName('dim_sorter')

        # make layout
        layout = QHBoxLayout()
        layout.addWidget(dim_sorter)
        pop.frame.setLayout(layout)

        # show popup
        pop.show_above_mouse()

    def _open_grid_popup(self):
        """Open grid options pop up widget."""

        # widgets
        popup = QtPopup(self)
        grid_stride = QtSpinBox(popup)
        grid_width = QtSpinBox(popup)
        grid_height = QtSpinBox(popup)
        grid_spacing = QDoubleSpinBox(popup)
        shape_help_symbol = QtToolTipLabel(self)
        stride_help_symbol = QtToolTipLabel(self)
        spacing_help_symbol = QtToolTipLabel(self)

        shape_help_msg = trans._(
            'Number of rows and columns in the grid. A value of -1 for either or both of width and height will trigger an auto calculation of the necessary grid shape to appropriately fill all the layers at the appropriate stride. 0 is not a valid entry.'
        )

        stride_help_msg = trans._(
            'Number of layers to place in each grid viewbox before moving on to the next viewbox. The default ordering is to place the most visible layer in the top left corner of the grid. A negative stride will cause the order in which the layers are placed in the grid to be reversed. 0 is not a valid entry.'
        )

        spacing_help_msg = trans._(
            'Spacing between grid viewboxes. If below 1, interpreted as percentage of the viewbox size, otherwise interpreted as screen pixels. 0 has the layers touching, positive values will space the grid positions apart.'
        )

        stride_min = self.viewer.grid.__fields__['stride'].type_.ge
        stride_max = self.viewer.grid.__fields__['stride'].type_.le
        stride_not = self.viewer.grid.__fields__['stride'].type_.ne
        grid_stride.setObjectName('gridStrideBox')
        grid_stride.setAlignment(Qt.AlignmentFlag.AlignCenter)
        grid_stride.setRange(stride_min, stride_max)
        grid_stride.setProhibitValue(stride_not)
        grid_stride.setValue(self.viewer.grid.stride)
        grid_stride.valueChanged.connect(self._update_grid_stride)
        self.grid_stride_box = grid_stride

        width_min = self.viewer.grid.__fields__['shape'].sub_fields[1].type_.ge
        width_not = self.viewer.grid.__fields__['shape'].sub_fields[1].type_.ne
        grid_width.setObjectName('gridWidthBox')
        grid_width.setAlignment(Qt.AlignmentFlag.AlignCenter)
        grid_width.setMinimum(width_min)
        grid_width.setProhibitValue(width_not)
        grid_width.setValue(self.viewer.grid.shape[1])
        grid_width.valueChanged.connect(self._update_grid_width)
        self.grid_width_box = grid_width

        height_min = (
            self.viewer.grid.__fields__['shape'].sub_fields[0].type_.ge
        )
        height_not = (
            self.viewer.grid.__fields__['shape'].sub_fields[0].type_.ne
        )
        grid_height.setObjectName('gridStrideBox')
        grid_height.setAlignment(Qt.AlignmentFlag.AlignCenter)
        grid_height.setMinimum(height_min)
        grid_height.setProhibitValue(height_not)
        grid_height.setValue(self.viewer.grid.shape[0])
        grid_height.valueChanged.connect(self._update_grid_height)
        self.grid_height_box = grid_height

        # set up spacing
        spacing_min = self.viewer.grid.__fields__['spacing'].type_.ge
        spacing_max = self.viewer.grid.__fields__['spacing'].type_.le
        spacing_step = self.viewer.grid.__fields__['spacing'].type_.step
        grid_spacing.setObjectName('gridSpacingBox')
        grid_spacing.setAlignment(Qt.AlignmentFlag.AlignCenter)
        grid_spacing.setMinimum(spacing_min)
        grid_spacing.setMaximum(spacing_max)
        grid_spacing.setValue(self.viewer.grid.spacing)
        grid_spacing.setDecimals(2)
        grid_spacing.setSingleStep(spacing_step)
        grid_spacing.valueChanged.connect(self._update_grid_spacing)
        self.grid_spacing_box = grid_spacing

        # help symbols
        shape_help_symbol.setObjectName('help_label')
        shape_help_symbol.setToolTip(shape_help_msg)

        stride_help_symbol.setObjectName('help_label')
        stride_help_symbol.setToolTip(stride_help_msg)

        spacing_help_symbol.setObjectName('help_label')
        spacing_help_symbol.setToolTip(spacing_help_msg)

        # layout
        grid_layout = QGridLayout()
        grid_layout.addWidget(QLabel(trans._('Grid stride:')), 0, 0)
        grid_layout.addWidget(grid_stride, 0, 1)
        grid_layout.addWidget(stride_help_symbol, 0, 2)

        grid_layout.addWidget(QLabel(trans._('Grid width:')), 1, 0)
        grid_layout.addWidget(grid_width, 1, 1)

        grid_layout.addWidget(shape_help_symbol, 1, 2, 2, 1)  # Span rows

        grid_layout.addWidget(QLabel(trans._('Grid height:')), 2, 0)
        grid_layout.addWidget(grid_height, 2, 1)

        grid_layout.addWidget(QLabel(trans._('Grid spacing:')), 3, 0)
        grid_layout.addWidget(grid_spacing, 3, 1)
        grid_layout.addWidget(spacing_help_symbol, 3, 2)

        popup.frame.setLayout(grid_layout)
        popup.show_above_mouse()

    def _update_grid_width(self, value):
        """Update the width value in grid shape.

        Parameters
        ----------
        value : int
            New grid width value.
        """

        self.viewer.grid.shape = (self.viewer.grid.shape[0], value)

    def _update_grid_stride(self, value):
        """Update stride in grid settings.

        Parameters
        ----------
        value : int
            New grid stride value.
        """

        self.viewer.grid.stride = value

    def _update_grid_height(self, value):
        """Update height value in grid shape.

        Parameters
        ----------
        value : int
            New grid height value.
        """

        self.viewer.grid.shape = (value, self.viewer.grid.shape[1])

    def _update_grid_spacing(self, value: float) -> None:
        """Update spacing value in grid settings.

        Parameters
        ----------
        value : float
            New grid spacing value.
        """
        self.viewer.grid.spacing = value


def _omit_viewer_args(constructor):
    @wraps(constructor)
    def _func(*args, **kwargs):
        if len(args) > 1 and not isinstance(args[1], str):
            warnings.warn(
                trans._(
                    'viewer argument is deprecated since 0.4.14 and should not be used'
                ),
                category=FutureWarning,
                stacklevel=2,
            )
            args = args[:1] + args[2:]
        if 'viewer' in kwargs:
            warnings.warn(
                trans._(
                    'viewer argument is deprecated since 0.4.14 and should not be used'
                ),
                category=FutureWarning,
                stacklevel=2,
            )
            del kwargs['viewer']
        return constructor(*args, **kwargs)

    return _func


class QtViewerPushButton(QPushButton):
    """Push button.

    Parameters
    ----------
    button_name : str
        Name of button.
    tooltip : str
        Tooltip for button. If empty then `button_name` is used
    slot : Callable, optional
        callable to be triggered on button click
    action : str
        action name to be triggered on button click
    """

    @_omit_viewer_args
    def __init__(
        self,
        button_name: str,
        tooltip: str = '',
        slot=None,
        action: str = '',
        extra_tooltip_text: str = '',
    ) -> None:
        super().__init__()

        self.setToolTip(tooltip or button_name)
        self.setProperty('mode', button_name)
        if slot is not None:
            self.clicked.connect(slot)
        if action:
            action_manager.bind_button(
                action, self, extra_tooltip_text=extra_tooltip_text
            )



import os
import re
import sys
import warnings
from collections.abc import Callable, Hashable, Iterator, Mapping
from typing import Any, Generic, TypeVar, Union

import wrapt

from napari.utils import misc
from napari.utils.translations import trans

_T = TypeVar('_T')
_K = TypeVar('_K', bound=Hashable)
_V = TypeVar('_V')


class ReadOnlyWrapper(wrapt.ObjectProxy):
    """
    Disable item and attribute setting with the exception of  ``__wrapped__``.
    """

    def __init__(self, wrapped: Any, exceptions: tuple[str, ...] = ()):
        super().__init__(wrapped)
        self._self_exceptions = exceptions

    def __setattr__(self, name: str, val: Any) -> None:
        if (
            name not in ('__wrapped__', '_self_exceptions')
            and name not in self._self_exceptions
        ):
            raise TypeError(
                trans._(
                    'cannot set attribute {name}',
                    deferred=True,
                    name=name,
                )
            )

        super().__setattr__(name, val)

    def __setitem__(self, name: str, val: Any) -> None:
        if name not in self._self_exceptions:
            raise TypeError(
                trans._('cannot set item {name}', deferred=True, name=name)
            )
        super().__setitem__(name, val)


_SUNDER = re.compile('^_[^_]')


class PublicOnlyProxy(wrapt.ObjectProxy, Generic[_T]):
    """Proxy to prevent private attribute and item access, recursively."""

    __wrapped__: _T

    @staticmethod
    def _is_private_attr(name: str) -> bool:
        return name.startswith('_') and not (
            name.startswith('__') and name.endswith('__')
        )

    @staticmethod
    def _private_attr_warning(name: str, typ: str) -> None:
        warnings.warn(
            trans._(
                "Private attribute access ('{typ}.{name}') in this context "
                '(e.g. inside a plugin widget or dock widget) is deprecated '
                'and will be unavailable in version 0.7.0',
                deferred=True,
                name=name,
                typ=typ,
            ),
            category=FutureWarning,
            stacklevel=3,
        )

        # This is code prepared for a moment where we want to block access to private attributes
        # raise AttributeError(
        #     trans._(
        #         "Private attribute set/access ('{typ}.{name}') not allowed in this context.",
        #         deferred=True,
        #         name=name,
        #         typ=typ,
        #     )
        # )

    @staticmethod
    def _is_called_from_napari() -> bool:
        """
        Check if the getter or setter is called from inner napari.
        """
        if hasattr(sys, '_getframe'):
            frame = sys._getframe(2)
            return frame.f_code.co_filename.startswith(misc.ROOT_DIR)
        return False

    def __getattr__(self, name: str) -> Any:
        if self._is_private_attr(name):
            # allow napari to access private attributes and get an non-proxy
            if self._is_called_from_napari():
                return super().__getattr__(name)

            typ = type(self.__wrapped__).__name__

            self._private_attr_warning(name, typ)
        with warnings.catch_warnings(record=True) as cx_manager:
            data = self.create(super().__getattr__(name))
        for warning in cx_manager:
            warnings.warn(warning.message, warning.category, stacklevel=2)

        return data

    def __setattr__(self, name: str, value: Any) -> None:
        if (
            os.environ.get('NAPARI_ENSURE_PLUGIN_MAIN_THREAD', '0')
            not in ('0', 'False')
        ) and not in_main_thread():
            raise RuntimeError(
                'Setting attributes on a napari object is only allowed from the main Qt thread.'
            )

        if self._is_private_attr(name):
            if self._is_called_from_napari():
                return super().__setattr__(name, value)

            typ = type(self.__wrapped__).__name__
            self._private_attr_warning(name, typ)

        if isinstance(value, PublicOnlyProxy):
            # if we want to set an attribute on a PublicOnlyProxy *and* the
            # value that we want to set is itself a PublicOnlyProxy, we unwrap
            # the value. This has two benefits:
            #
            # 1. Checking the attribute later will incur a significant
            # performance cost, because _is_called_from_napari() will be
            # checked on each attribute access and it involves inspecting the
            # calling frame, which is expensive.
            # 2. Certain equality checks fail when objects are
            # PublicOnlyProxies. Notably, equality checks fail when such
            # objects are included in a Qt data model. For example, plugins can
            # grab a layer from the viewer; this layer will be wrapped by the
            # PublicOnlyProxy, and then using this object to set the current
            # layer selection will not propagate the selection to the Viewer.
            # See https://github.com/napari/napari/issues/5767
            value = value.__wrapped__

        setattr(self.__wrapped__, name, value)
        return None

    def __getitem__(self, key: Any) -> Any:
        return self.create(super().__getitem__(key))

    def __repr__(self) -> str:
        return repr(self.__wrapped__)

    def __dir__(self) -> list[str]:
        return [x for x in dir(self.__wrapped__) if not _SUNDER.match(x)]

    @classmethod
    def create(cls, obj: Any) -> Union['PublicOnlyProxy', Any]:
        # restrict the scope of this proxy to napari objects
        if type(obj).__name__ == 'method':
            # If the given object is a method, we check the module *of the
            # object to which that method is bound*. Otherwise, the module of a
            # method is just builtins!
            mod = getattr(type(obj.__self__), '__module__', None) or ''
        else:
            # Otherwise, the module is of an object just given by the
            # __module__ attribute.
            mod = getattr(type(obj), '__module__', None) or ''
        if not mod.startswith('napari'):
            return obj
        if isinstance(obj, PublicOnlyProxy):
            return obj  # don't double-wrap
        if callable(obj):
            return CallablePublicOnlyProxy(obj)
        return PublicOnlyProxy(obj)


class CallablePublicOnlyProxy(PublicOnlyProxy[Callable]):
    def __call__(self, *args, **kwargs):  # type: ignore [no-untyped-def]
        # if a PublicOnlyProxy is callable, then when we call it we:
        # - unwrap the arguments, to avoid performance issues detailed in
        #   PublicOnlyProxy.__setattr__,
        # - call the unwrapped callable on the unwrapped arguments
        # - wrap the result in a PublicOnlyProxy
        args = tuple(
            arg.__wrapped__ if isinstance(arg, PublicOnlyProxy) else arg
            for arg in args
        )
        kwargs = {
            k: v.__wrapped__ if isinstance(v, PublicOnlyProxy) else v
            for k, v in kwargs.items()
        }
        return self.create(self.__wrapped__(*args, **kwargs))


class MappingProxy(Mapping[_K, _V]):
    def __init__(self, wrapped: Mapping[_K, _V]) -> None:
        """
        A read-only proxy for a mapping object.

        Parameters
        ----------
        wrapped : Mapping
            The mapping to wrap.
        """
        self._wrapped = wrapped

    def __getitem__(self, key: _K, /) -> _V:
        """
        Get an item from the wrapped mapping.

        Parameters
        ----------
        key : Any
            The key to get from the mapping.

        Returns
        -------
        value : Any
            The value associated with the key in the wrapped mapping.
        """
        return self._wrapped[key]

    def __len__(self) -> int:
        """
        Get the length of the wrapped mapping.

        Returns
        -------
        length : int
            The number of items in the wrapped mapping.
        """
        return len(self._wrapped)

    def __iter__(self) -> Iterator[_K]:
        """
        Get an iterator over the keys of the wrapped mapping.

        Returns
        -------
        iterator : Iterator
            An iterator over the keys of the wrapped mapping.
        """
        return iter(self._wrapped)

    def __contains__(self, item: Hashable) -> bool:
        """
        Check if the wrapped mapping contains a key.

        Parameters
        ----------
        item : Any
            The key to check for in the wrapped mapping.

        Returns
        -------
        contains : bool
            True if the key is in the wrapped mapping, False otherwise.
        """
        return item in self._wrapped


def in_main_thread_py() -> bool:
    """
    Check if caller is in main python thread.

    Returns
    -------
    thread_flag : bool
        True if we are in the main thread, False otherwise.
    """
    import threading

    return threading.current_thread() == threading.main_thread()


def _in_main_thread() -> bool:
    """
    General implementation of checking if we are in a proper thread.
    If Qt is available and Application is created then assign :py:func:`in_qt_main_thread` to `in_main_thread`.
    If Qt liba are not available then assign :py:func:`in_main_thread_py` to in_main_thread.
    IF Qt libs are available but there is no Application ti wil emmit warning and return result of in_main_thread_py.

    Returns
    -------
    thread_flag : bool
        True if we are in the main thread, False otherwise.
    """

    global in_main_thread
    try:
        from napari._qt.utils import in_qt_main_thread

        res = in_qt_main_thread()
        in_main_thread = in_qt_main_thread
    except ImportError:
        in_main_thread = in_main_thread_py
        return in_main_thread_py()
    except AttributeError:
        warnings.warn(
            'Qt libs are available but no QtApplication instance is created'
        )
        return in_main_thread_py()
    return res


in_main_thread = _in_main_thread



from enum import auto
from typing import Literal

from napari.utils.misc import StringEnum


class VerticalAxisOrientation(StringEnum):
    UP = auto()
    DOWN = auto()


class HorizontalAxisOrientation(StringEnum):
    LEFT = auto()
    RIGHT = auto()


class DepthAxisOrientation(StringEnum):
    AWAY = auto()
    TOWARDS = auto()


class Handedness(StringEnum):
    RIGHT = auto()
    LEFT = auto()


VerticalAxisOrientationStr = Literal['up', 'down']
HorizontalAxisOrientationStr = Literal['left', 'right']
DepthAxisOrientationStr = Literal['away', 'towards']

# Prior to v0.6.0, the default would be equivalent to ('away', 'down', 'right')
DEFAULT_ORIENTATION_TYPED = (
    DepthAxisOrientation.TOWARDS,
    VerticalAxisOrientation.DOWN,
    HorizontalAxisOrientation.RIGHT,
)
DEFAULT_ORIENTATION = tuple(map(str, DEFAULT_ORIENTATION_TYPED))



"""RemoteManager class."""

import logging

from napari.components.experimental.remote._commands import RemoteCommands
from napari.components.experimental.remote._messages import RemoteMessages
from napari.components.layerlist import LayerList
from napari.utils.events import Event

LOGGER = logging.getLogger('napari.monitor')


class RemoteManager:
    """Interacts with remote clients.

    The monitor system itself purposely does not depend on anything else in
    napari except for utils.events.

    However RemoteManager and its children RemoteCommands and
    RemoteMessages do very much depend on napari. RemoteCommands executes
    commands sent to napari by clients. RemoteMessages sends messages to
    remote clients, such as the current state of the layers.

    Parameters
    ----------
    layers : LayerList
        The viewer's layers.
    """

    def __init__(self, layers: LayerList) -> None:
        self._commands = RemoteCommands(layers)
        self._messages = RemoteMessages(layers)

    def process_command(self, event: Event) -> None:
        """Process this command from a remote client.

        Parameters
        ----------
        event : Event
            Contains the command to process.
        """
        return self._commands.process_command(event)

    def on_poll(self, _event: Event) -> None:
        """Send out messages when polled."""
        self._messages.on_poll()



import sys
from itertools import product

import numpy as np
import pytest


@pytest.mark.skipif(
    sys.platform == 'win32', reason='This new test is flaky on windows'
)
def test_bounding_box_multiscale_3D(make_napari_viewer, qtbot):
    viewer = make_napari_viewer(show=True)

    data = np.ones((2, 200, 200))
    layer = viewer.add_image(
        [data, data[:, ::2, ::2], data[:, ::4, ::4]], multiscale=True
    )
    viewer.layers[-1].bounding_box.visible = True
    assert viewer.dims.ndisplay == 2

    # Set canvas size to target amount
    viewer.window._qt_viewer.canvas.size = (200, 200)
    viewer.window._qt_viewer.canvas.on_draw(None)
    viewer.camera.zoom = 2

    assert viewer.layers[0].data_level == 0

    viewer.dims.ndisplay = 3

    qtbot.waitUntil(lambda: viewer.layers[-1]._loaded)
    # get the actual bounding box vertices
    bb = viewer.window._qt_viewer.canvas._layer_overlay_to_visual[layer][
        layer.bounding_box
    ]
    displayed_bbox_vertices = bb.node.markers._data['a_position'].astype(
        'float'
    )

    # for multiscale layers, in 3D, the lowest data level is displayed
    # get the vertices from the lowest data level bounding box, augmented
    expected_vertices = np.array(
        list(
            product(
                *viewer.layers[-1]._display_bounding_box_at_level(
                    viewer.dims.displayed, len(viewer.layers[-1].data) - 1
                )
                + np.array([[-0.5, 0.5]])
            )
        )
    )
    # roll the coordinates to match the vispy marker vertices
    expected_vertices = np.roll(expected_vertices, shift=-1, axis=1).astype(
        'float'
    )

    # the order of the vertices is not important, just the locations
    assert set(map(tuple, displayed_bbox_vertices)) == set(
        map(tuple, expected_vertices)
    )



import itertools

import numpy as np
import pytest
import zarr
from dask import array as da

from napari.utils._dtype import get_dtype_limits, normalize_dtype

ts = pytest.importorskip('tensorstore')
torch = pytest.importorskip('torch')

bit_depths = [str(2**i) for i in range(3, 7)]
uints = ['uint' + b for b in bit_depths]
ints = ['int' + b for b in bit_depths]
floats = ['float32', 'float64']
complex_types = ['complex64', 'complex128']
bools = ['bool']
pure_py = ['int', 'float']


@pytest.mark.parametrize(
    'dtype_str', ['uint8', *ints, *floats, *complex_types, *bools]
)
def test_normalize_dtype_torch(dtype_str):
    """torch doesn't have uint for >8bit, so it gets its own test."""
    # torch doesn't let you specify dtypes as str,
    # see https://github.com/pytorch/pytorch/issues/40568
    torch_arr = torch.zeros(5, dtype=getattr(torch, dtype_str))
    np_arr = np.zeros(5, dtype=dtype_str)
    assert normalize_dtype(torch_arr.dtype) == np.dtype(np_arr.dtype)


@pytest.mark.parametrize(
    'dtype_str', uints + ints + floats + complex_types + bools
)
def test_normalize_dtype_tensorstore(dtype_str):
    np_arr = np.zeros(5, dtype=dtype_str)
    ts_arr = ts.array(np_arr)  # inherit ts dtype from np dtype
    assert normalize_dtype(ts_arr.dtype) is np.dtype(np_arr.dtype)


@pytest.mark.parametrize(
    ('module', 'dtype_str'),
    itertools.product(
        (np, da, zarr), uints + ints + floats + complex_types + bools
    ),
)
def test_normalize_dtype_np_noop(module, dtype_str):
    """Check that normalize dtype works as expected for plain NumPy dtypes."""
    module_arr = module.zeros(5, dtype=dtype_str)
    np_arr = np.zeros(5, dtype=normalize_dtype(module_arr.dtype))
    assert normalize_dtype(module_arr.dtype) is normalize_dtype(np_arr.dtype)


@pytest.mark.parametrize('dtype_str', ['int', 'float'])
def test_pure_python_types(dtype_str):
    pure_arr = np.zeros(5, dtype=dtype_str)
    norm_arr = np.zeros(5, dtype=normalize_dtype(dtype_str))
    assert pure_arr.dtype is norm_arr.dtype


# note: we don't write specific tests for zarr and dask because they use numpy
# dtypes directly.


@pytest.mark.parametrize(
    'dtype',
    [
        int,
        'uint8',
        np.uint8,
        'int8',
        'uint16',
        'int16',
        'uint32',
        'int32',
        float,
        'float32',
        'float64',
        '>f4',
        '>f8',
    ]
    + [''.join(t) for t in itertools.product('<>', 'iu', '1248')],
)
def test_dtype_lims(dtype):
    lims = get_dtype_limits(dtype)
    assert isinstance(lims, tuple)
    assert len(lims) == 2



import numpy as np
from vispy.scene.visuals import Compound, Line, Mesh, Text

from napari.layers.shapes._shapes_utils import triangulate_ellipse
from napari.utils.colormaps.standardize_color import transform_color
from napari.utils.translations import trans


def make_dashed_line(num_dashes, axis):
    """Make a dashed line.

    Parameters
    ----------
    num_dashes : int
        Number of dashes in the line.
    axis : int
        Axis which is dashed.

    Returns
    -------
    np.ndarray
        Dashed line, of shape (num_dashes, 3) with zeros in
        the non dashed axes and line segments in the dashed
        axis.
    """
    dashes = np.linspace(0, 1, num_dashes * 2)
    dashed_line_ends = np.concatenate(
        [[dashes[2 * i], dashes[2 * i + 1]] for i in range(num_dashes)], axis=0
    )
    dashed_line = np.zeros((2 * num_dashes, 3))
    dashed_line[:, axis] = np.array(dashed_line_ends)
    return dashed_line


def make_arrow_head(num_segments, axis):
    """Make an arrowhead line.

    Parameters
    ----------
    num_segments : int
        Number of segments in the arrowhead.
    axis
        Arrowhead direction.

    Returns
    -------
    np.ndarray, np.ndarray
        Vertices and faces of the arrowhead.
    """
    corners = np.array([[-1, -1], [-1, 1], [1, 1], [1, -1]]) * 0.1
    vertices, faces = triangulate_ellipse(corners, num_segments)  # type: ignore[arg-type]
    full_vertices = np.zeros((num_segments + 1, 3))
    inds = list(range(3))
    inds.pop(axis)
    full_vertices[:, inds] = vertices
    full_vertices[:, axis] = 0.9
    full_vertices[0, axis] = 1.02
    return full_vertices, faces


def color_lines(colors):
    if len(colors) == 2:
        return np.concatenate(
            [[colors[0]] * 2, [colors[1]] * 2],
            axis=0,
        )
    if len(colors) == 3:
        return np.concatenate(
            [[colors[0]] * 2, [colors[1]] * 2, [colors[2]] * 2],
            axis=0,
        )

    return ValueError(
        trans._(
            'Either 2 or 3 colors must be provided, got {number}.',
            deferred=True,
            number=len(colors),
        )
    )


def color_dashed_lines(colors):
    if len(colors) == 2:
        return np.concatenate(
            [[colors[0]] * 2, [colors[1]] * 4 * 2],
            axis=0,
        )
    if len(colors) == 3:
        return np.concatenate(
            [[colors[0]] * 2, [colors[1]] * 4 * 2, [colors[2]] * 8 * 2],
            axis=0,
        )

    return ValueError(
        trans._(
            'Either 2 or 3 colors must be provided, got {number}.',
            deferred=True,
            number=len(colors),
        )
    )


def color_arrowheads(colors, num_segments):
    if len(colors) == 2:
        return np.concatenate(
            [[colors[0]] * num_segments, [colors[1]] * num_segments],
            axis=0,
        )
    if len(colors) == 3:
        return np.concatenate(
            [
                [colors[0]] * num_segments,
                [colors[1]] * num_segments,
                [colors[2]] * num_segments,
            ],
            axis=0,
        )

    return ValueError(
        trans._(
            'Either 2 or 3 colors must be provided, got {number}.',
            deferred=True,
            number=len(colors),
        )
    )


class Axes(Compound):
    def __init__(self) -> None:
        self._num_segments_arrowhead = 100
        # CMYRGB for 6 axes data in x, y, z, ... ordering
        self._default_color = [
            [0, 1, 1, 1],
            [1, 0, 1, 1],
            [1, 1, 0, 1],
            [1, 0, 0, 1],
            [0, 1, 0, 1],
            [0, 0, 1, 1],
        ]

        self._text_offsets = 0.1 * np.array([1, 1, 1])

        # note order is x, y, z for VisPy
        self._line_data2D = np.array(
            [[0, 0, 0], [1, 0, 0], [0, 0, 0], [0, 1, 0]]
        )
        self._line_data3D = np.array(
            [[0, 0, 0], [1, 0, 0], [0, 0, 0], [0, 1, 0], [0, 0, 0], [0, 0, 1]]
        )

        # note order is x, y, z for VisPy
        self._dashed_line_data2D = np.concatenate(
            [[[1, 0, 0], [0, 0, 0]], make_dashed_line(4, axis=1)],
            axis=0,
        )
        self._dashed_line_data3D = np.concatenate(
            [
                [[1, 0, 0], [0, 0, 0]],
                make_dashed_line(4, axis=1),
                make_dashed_line(8, axis=2),
            ],
            axis=0,
        )

        # note order is x, y, z for VisPy
        vertices = np.empty((0, 3))
        faces = np.empty((0, 3))
        for axis in range(2):
            v, f = make_arrow_head(self._num_segments_arrowhead, axis)
            faces = np.concatenate([faces, f + len(vertices)], axis=0)
            vertices = np.concatenate([vertices, v], axis=0)
        self._default_arrow_vertices2D = vertices
        self._default_arrow_faces2D = faces.astype(int)

        vertices = np.empty((0, 3))
        faces = np.empty((0, 3))
        for axis in range(3):
            v, f = make_arrow_head(self._num_segments_arrowhead, axis)
            faces = np.concatenate([faces, f + len(vertices)], axis=0)
            vertices = np.concatenate([vertices, v], axis=0)
        self._default_arrow_vertices3D = vertices
        self._default_arrow_faces3D = faces.astype(int)

        super().__init__(
            [
                Line(connect='segments', method='gl', width=3),
                Mesh(),
                Text(
                    text='1',
                    font_size=10,
                    anchor_x='center',
                    anchor_y='center',
                ),
            ]
        )

    @property
    def line(self):
        return self._subvisuals[0]

    @property
    def mesh(self):
        return self._subvisuals[1]

    @property
    def text(self):
        return self._subvisuals[2]

    def set_data(self, axes, reversed_axes, colored, bg_color, dashed, arrows):
        ndisplay = len(axes)

        # Determine colors of axes based on reverse position
        if colored:
            axes_colors = [
                self._default_color[ra % len(self._default_color)]
                for ra in reversed_axes
            ]
        else:
            # the reason for using the `as_hex` here is to avoid
            # `UserWarning` which is emitted when RGB values are above 1
            bg_color = transform_color(bg_color.as_hex())[0]
            color = np.subtract(1, bg_color)
            color[-1] = bg_color[-1]
            axes_colors = [color] * ndisplay

        # Determine data based on number of displayed dimensions and
        # axes visualization parameters
        if dashed and ndisplay == 2:
            data = self._dashed_line_data2D
            color = color_dashed_lines(axes_colors)
            text_data = self._line_data2D[1::2]
        elif dashed and ndisplay == 3:
            data = self._dashed_line_data3D
            color = color_dashed_lines(axes_colors)
            text_data = self._line_data3D[1::2]
        elif not dashed and ndisplay == 2:
            data = self._line_data2D
            color = color_lines(axes_colors)
            text_data = self._line_data2D[1::2]
        elif not dashed and ndisplay == 3:
            data = self._line_data3D
            color = color_lines(axes_colors)
            text_data = self._line_data3D[1::2]
        else:
            raise ValueError(
                trans._(
                    'Axes dash status and ndisplay combination not supported',
                    deferred=True,
                )
            )

        if arrows and ndisplay == 2:
            arrow_vertices = self._default_arrow_vertices2D
            arrow_faces = self._default_arrow_faces2D
            arrow_color = color_arrowheads(
                axes_colors, self._num_segments_arrowhead
            )
        elif arrows and ndisplay == 3:
            arrow_vertices = self._default_arrow_vertices3D
            arrow_faces = self._default_arrow_faces3D
            arrow_color = color_arrowheads(
                axes_colors, self._num_segments_arrowhead
            )
        else:
            arrow_vertices = np.zeros((3, 3))
            arrow_faces = np.array([[0, 1, 2]])
            arrow_color = [[0, 0, 0, 0]]

        self.line.set_data(data, color)
        self.mesh.set_data(
            vertices=arrow_vertices,
            faces=arrow_faces,
            face_colors=arrow_color,
        )

        self.text.color = axes_colors
        self.text.pos = text_data + self._text_offsets



import os
from unittest import mock

import numpy as np
import pytest
import zarr
from npe2 import DynamicPlugin
from npe2.manifest.contributions import SampleDataURI
from qtpy.QtWidgets import QLabel, QRadioButton

from napari._app_model import get_app_model
from napari._qt.dialogs.qt_reader_dialog import (
    QtReaderDialog,
    open_with_dialog_choices,
    prepare_remaining_readers,
)
from napari._qt.qt_viewer import QtViewer
from napari.components import ViewerModel
from napari.errors.reader_errors import ReaderPluginError
from napari.settings import get_settings


@pytest.fixture
def reader_dialog(qtbot):
    def _reader_dialog(**kwargs):
        widget = QtReaderDialog(**kwargs)
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _reader_dialog


def test_reader_dialog_buttons(reader_dialog):
    widg = reader_dialog(
        readers={'display name': 'plugin-name', 'display 2': 'plugin2'}
    )
    assert len(widg.findChildren(QRadioButton)) == 2


def test_reader_defaults(reader_dialog, tmpdir):
    file_pth = tmpdir.join('my_file.tif')
    widg = reader_dialog(pth=file_pth, readers={'p1': 'p1', 'p2': 'p2'})

    assert widg.findChild(QLabel).text().startswith('Choose reader')
    assert widg._get_plugin_choice() == 'p1'
    assert not widg.persist_checkbox.isChecked()


def test_reader_with_error_message(reader_dialog):
    widg = reader_dialog(error_message='Test Error')
    assert widg.findChild(QLabel).text().startswith('Test Error')


def test_reader_dir_with_extension(tmpdir, reader_dialog):
    dir_name = tmpdir.mkdir('my_dir.zarr')
    widg = reader_dialog(pth=dir_name, readers={'p1': 'p1', 'p2': 'p2'})
    assert hasattr(widg, 'persist_checkbox')
    assert (
        widg.persist_checkbox.text()
        == 'Remember this choice for files with a .zarr extension'
    )


def test_reader_dir(tmpdir, reader_dialog):
    dir_name = tmpdir.mkdir('my_dir')
    widg = reader_dialog(pth=dir_name, readers={'p1': 'p1', 'p2': 'p2'})
    assert (
        widg._persist_text
        == f'Remember this choice for folders labeled as {dir_name}{os.sep}.'
    )


def test_get_plugin_choice(tmpdir, reader_dialog):
    file_pth = tmpdir.join('my_file.tif')
    widg = reader_dialog(pth=file_pth, readers={'p1': 'p1', 'p2': 'p2'})
    reader_btns = widg.reader_btn_group.buttons()

    reader_btns[1].toggle()
    assert widg._get_plugin_choice() == 'p2'

    reader_btns[0].toggle()
    assert widg._get_plugin_choice() == 'p1'


def test_get_persist_choice(tmpdir, reader_dialog):
    file_pth = tmpdir.join('my_file.tif')
    widg = reader_dialog(pth=file_pth, readers={'p1': 'p1', 'p2': 'p2'})
    assert not widg._get_persist_choice()

    widg.persist_checkbox.toggle()
    assert widg._get_persist_choice()


def test_prepare_dialog_options_no_readers():
    with pytest.raises(ReaderPluginError) as e:
        prepare_remaining_readers(
            ['my-file.fake'], 'fake-reader', RuntimeError('Reading failed')
        )
    assert 'Tried to read my-file.fake with plugin fake-reader' in str(e.value)


def test_prepare_dialog_options_multiple_plugins(builtins):
    pth = 'my-file.tif'

    readers = prepare_remaining_readers(
        [pth],
        None,
        RuntimeError(f'Multiple plugins found capable of reading {pth}'),
    )
    assert builtins.name in readers


def test_prepare_dialog_options_removes_plugin(tmp_plugin: DynamicPlugin):
    tmp2 = tmp_plugin.spawn(register=True)

    @tmp_plugin.contribute.reader(filename_patterns=['*.fake'])
    def _(path): ...

    @tmp2.contribute.reader(filename_patterns=['*.fake'])
    def _(path): ...

    readers = prepare_remaining_readers(
        ['my-file.fake'],
        tmp_plugin.name,
        RuntimeError('Reader failed'),
    )
    assert tmp2.name in readers
    assert tmp_plugin.name not in readers


def test_open_sample_data_shows_all_readers(
    make_napari_viewer,
    tmp_plugin: DynamicPlugin,
):
    """Checks that sample data callback `_add_sample` shows all readers."""
    # Test for bug fixed in #6058
    tmp2 = tmp_plugin.spawn(register=True)

    @tmp_plugin.contribute.reader(filename_patterns=['*.fake'])
    def _(path): ...

    @tmp2.contribute.reader(filename_patterns=['*.fake'])
    def _(path): ...

    my_sample = SampleDataURI(
        key='tmp-sample',
        display_name='Temp Sample',
        uri='some-path/some-file.fake',
    )
    tmp_plugin.manifest.contributions.sample_data = [my_sample]

    app = get_app_model()
    # required so setup steps run in init of `Viewer` and `Window`
    viewer = make_napari_viewer()
    # Ensure that `handle_gui_reading`` is not passed the sample plugin name
    with mock.patch(
        'napari._qt.dialogs.qt_reader_dialog.handle_gui_reading'
    ) as mock_read:
        app.commands.execute_command('tmp_plugin:tmp-sample')

    mock_read.assert_called_once_with(
        ['some-path/some-file.fake'],
        viewer.window._qt_viewer,
        stack=False,
    )


def test_open_with_dialog_choices_persist(builtins, tmp_path, qtbot):
    pth = tmp_path / 'my-file.npy'
    np.save(pth, np.random.random((10, 10)))

    viewer = ViewerModel()
    qt_viewer = QtViewer(viewer)
    qtbot.addWidget(qt_viewer)

    open_with_dialog_choices(
        display_name=builtins.display_name,
        persist=True,
        extension='.npy',
        readers={builtins.name: builtins.display_name},
        paths=[str(pth)],
        stack=False,
        qt_viewer=qt_viewer,
    )
    assert len(viewer.layers) == 1
    # make sure extension was saved with *
    assert get_settings().plugins.extension2reader['*.npy'] == builtins.name


def test_open_with_dialog_choices_persist_dir(builtins, tmp_path, qtbot):
    pth = tmp_path / 'data.zarr'
    z = zarr.open(
        store=str(pth), mode='w', shape=(10, 10), chunks=(5, 5), dtype='f4'
    )
    z[:] = np.random.random((10, 10))

    viewer = ViewerModel()
    qt_viewer = QtViewer(viewer)
    qtbot.addWidget(qt_viewer)

    open_with_dialog_choices(
        display_name=builtins.display_name,
        persist=True,
        extension=str(pth),
        readers={builtins.name: builtins.display_name},
        paths=[str(pth)],
        stack=False,
        qt_viewer=qt_viewer,
    )
    assert len(viewer.layers) == 1
    # make sure extension was saved without * and with trailing slash
    assert (
        get_settings().plugins.extension2reader[f'{pth}{os.sep}']
        == builtins.name
    )


def test_open_with_dialog_choices_raises(make_napari_viewer):
    viewer = make_napari_viewer()

    get_settings().plugins.extension2reader = {}
    with pytest.raises(ValueError, match='does not exist'):
        open_with_dialog_choices(
            display_name='Fake Plugin',
            persist=True,
            extension='.fake',
            readers={'fake-plugin': 'Fake Plugin'},
            paths=['my-file.fake'],
            stack=False,
            qt_viewer=viewer.window._qt_viewer,
        )
    # settings weren't saved because reading failed
    assert not get_settings().plugins.extension2reader



import gc
import os
import sys
import warnings
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import patch
from weakref import WeakSet

import pytest

if TYPE_CHECKING:
    from pytest import FixtureRequest  # noqa: PT013

_SAVE_GRAPH_OPNAME = '--save-leaked-object-graph'


def _empty(*_, **__):
    """Empty function for mocking"""


def pytest_addoption(parser):
    parser.addoption(
        '--show-napari-viewer',
        action='store_true',
        default=False,
        help="don't show viewer during tests",
    )

    parser.addoption(
        _SAVE_GRAPH_OPNAME,
        action='store_true',
        default=False,
        help="Try to save a graph of leaked object's reference (need objgraph"
        'and graphviz installed',
    )


COUNTER = 0


def fail_obj_graph(Klass):  # pragma: no cover
    """
    Fail is a given class _instances weakset is non empty and print the object graph.
    """

    try:
        import objgraph
    except ModuleNotFoundError:
        return

    if len(Klass._instances) != 0:
        global COUNTER
        COUNTER += 1
        import gc

        leaked_objects_count = len(Klass._instances)

        gc.collect()
        file_path = Path(
            f'{Klass.__name__}-leak-backref-graph-{COUNTER}.pdf'
        ).absolute()
        objgraph.show_backrefs(
            list(Klass._instances),
            max_depth=20,
            filename=str(file_path),
        )

        Klass._instances.clear()

        assert file_path.exists()

        # DO not remove len, this can break as C++ obj are gone, but python objects
        # still hang around and _repr_ would crash.
        pytest.fail(
            f'Test run fail with leaked {leaked_objects_count} instances of {Klass}.'
            f'The object graph is saved in {file_path}.'
            f'{len(Klass._instances)} objects left after cleanup'
        )


@pytest.fixture
def napari_plugin_manager(monkeypatch):
    """A napari plugin manager that blocks discovery by default.

    Note you can still use `napari_plugin_manager.register()` to directly
    register a plugin module, class or dict for testing.

    Or, to re-enable global discovery, use:
    `napari_plugin_manager.discovery_blocker.stop()`
    """
    import napari
    import napari.plugins.io
    from napari.plugins._plugin_manager import NapariPluginManager

    pm = NapariPluginManager()

    # make it so that internal requests for the plugin_manager
    # get this test version for the duration of the test.
    monkeypatch.setattr(napari.plugins, 'plugin_manager', pm)
    monkeypatch.setattr(napari.plugins.io, 'plugin_manager', pm)
    with suppress(AttributeError):
        monkeypatch.setattr(napari._qt.qt_main_window, 'plugin_manager', pm)
    # prevent discovery of plugins in the environment
    # you can still use `pm.register` to explicitly register something.
    pm.discovery_blocker = patch.object(pm, 'discover')
    pm.discovery_blocker.start()
    pm._initialize()  # register our builtins
    yield pm
    pm.discovery_blocker.stop()


GCPASS = 0


@pytest.fixture(autouse=True)
def _clean_themes():
    from napari.utils import theme

    themes = set(theme.available_themes())
    yield
    for name in theme.available_themes():
        if name not in themes:
            del theme._themes[name]


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # https://docs.pytest.org/en/latest/example/simple.html#making-test-result-information-available-in-fixtures
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"

    setattr(item, f'rep_{rep.when}', rep)


@pytest.fixture
def mock_app_model():
    """Mock clean 'test_app' `NapariApplication` instance.

    This fixture must be used whenever `napari._app_model.get_app_model()` is called to
    return a 'test_app' `NapariApplication` instead of the 'napari'
    `NapariApplication`. The `make_napari_viewer` fixture is already equipped with
    a `mock_app_model`.

    Note that `NapariApplication` registers app-model actions.
    If this is not desired, please create a clean
    `app_model.Application` in the test.

    It does not register Qt related actions, providers or processors, which is done
    via `init_qactions()`. Nor does it register plugins, done via `_initialize_plugins`.
    If these are required, the `make_napari_viewer` fixture can be used, which
    will register ALL actions, providers and processors and register plugins.
    It will also automatically clear the lru cache.

    Alternatively, you can specifically run `init_qactions()` or
    `_initialize_plugins` within the test, ensuring that you `cache_clear()`
    first.
    """
    from app_model import Application

    from napari._app_model._app import NapariApplication, _napari_names

    app = NapariApplication('test_app')
    app.injection_store.namespace = _napari_names
    with patch.object(NapariApplication, 'get_app_model', return_value=app):
        try:
            yield app
        finally:
            Application.destroy('test_app')


@pytest.fixture
def make_napari_viewer(
    qtbot,
    request: 'FixtureRequest',
    mock_app_model,
    napari_plugin_manager,
    monkeypatch,
):
    """A pytest fixture function that creates a napari viewer for use in testing.

    This fixture will take care of creating a viewer and cleaning up at the end of the
    test. When using this function, it is **not** necessary to use a `qtbot` fixture,
    nor should you do any additional cleanup (such as using `qtbot.addWidget` or
    calling `viewer.close()`) at the end of the test. Duplicate cleanup may cause
    an error.

    To use this fixture as a function in your tests:

        def test_something_with_a_viewer(make_napari_viewer):
            # `make_napari_viewer` takes any keyword arguments that napari.Viewer() takes
            viewer = make_napari_viewer()

    It accepts all the same arguments as `napari.Viewer`, notably `show`
    which should be set to `True` for tests that require the `Viewer` to be visible
    (e.g., tests that check aspects of the Qt window or layer rendering).
    It also accepts the following test-related parameters:

    ViewerClass : Type[napari.Viewer], optional
        Override the viewer class being used.  By default, will
        use napari.viewer.Viewer
    strict_qt : bool or str, optional
        If True, a check will be performed after test cleanup to make sure that
        no top level widgets were created and *not* cleaned up during the
        test.  If the string "raise" is provided, an AssertionError will be
        raised.  Otherwise a warning is emitted.
        By default, this is False unless the test is being performed within
        the napari package.
        This can be made globally true by setting the 'NAPARI_STRICT_QT'
        environment variable.
    block_plugin_discovery : bool, optional
        Block discovery of non-builtin plugins.  Note: plugins can still be
        manually registered by using the 'napari_plugin_manager' fixture and
        the `napari_plugin_manager.register()` method. By default, True.

    Examples
    --------
    >>> def test_adding_shapes(make_napari_viewer):
    ...     viewer = make_napari_viewer()
    ...     viewer.add_shapes()
    ...     assert len(viewer.layers) == 1

    >>> def test_something_with_plugins(make_napari_viewer):
    ...     viewer = make_napari_viewer(block_plugin_discovery=False)

    >>> def test_something_with_strict_qt_tests(make_napari_viewer):
    ...     viewer = make_napari_viewer(strict_qt=True)
    """
    from qtpy.QtWidgets import QApplication, QWidget

    from napari import Viewer
    from napari._qt._qapp_model.qactions import init_qactions
    from napari._qt.qt_viewer import QtViewer
    from napari.plugins import _initialize_plugins
    from napari.settings import get_settings

    global GCPASS
    GCPASS += 1

    if GCPASS % 50 == 0:
        gc.collect()
    else:
        gc.collect(1)

    _do_not_inline_below = len(QtViewer._instances)
    # # do not inline to avoid pytest trying to compute repr of expression.
    # # it fails if C++ object gone but not Python object.
    if request.config.getoption(_SAVE_GRAPH_OPNAME):
        fail_obj_graph(QtViewer)
    QtViewer._instances.clear()
    assert _do_not_inline_below == 0, (
        'Some instance of QtViewer is not properly cleaned in one of previous test. For easier debug one may '
        f'use {_SAVE_GRAPH_OPNAME} flag for pytest to get graph of leaked objects. If you use qtbot (from pytest-qt)'
        ' to clean Qt objects after test you may need to switch to manual clean using '
        '`deleteLater()` and `qtbot.wait(50)` later.'
    )

    settings = get_settings()
    settings.reset()

    _initialize_plugins.cache_clear()
    init_qactions.cache_clear()

    viewers: WeakSet[Viewer] = WeakSet()
    request.node._viewer_weak_set = viewers

    # may be overridden by using the parameter `strict_qt`
    _strict = False

    initial = QApplication.topLevelWidgets()
    prior_exception = getattr(sys, 'last_value', None)
    is_internal_test = request.module.__name__.startswith('napari.')

    # disable thread for status checker
    monkeypatch.setattr(
        'napari._qt.threads.status_checker.StatusChecker.start',
        _empty,
    )

    if 'enable_console' not in request.keywords:

        def _dummy_widget(*_):
            w = QWidget()
            w._update_theme = _empty
            return w

        monkeypatch.setattr(
            'napari._qt.qt_viewer.QtViewer._get_console', _dummy_widget
        )

    def actual_factory(
        *model_args,
        ViewerClass=Viewer,
        strict_qt=None,
        block_plugin_discovery=True,
        **model_kwargs,
    ):
        if strict_qt is None:
            strict_qt = is_internal_test or os.getenv('NAPARI_STRICT_QT')
        nonlocal _strict
        _strict = strict_qt

        if not block_plugin_discovery:
            napari_plugin_manager.discovery_blocker.stop()

        should_show = request.config.getoption('--show-napari-viewer')
        model_kwargs['show'] = model_kwargs.pop('show', should_show)
        viewer = ViewerClass(*model_args, **model_kwargs)
        viewers.add(viewer)

        return viewer

    yield actual_factory

    # Some tests might have the viewer closed, so this call will not be able
    # to access the window.
    with suppress(AttributeError):
        get_settings().reset()

    # close viewers, but don't saving window settings while closing
    for viewer in viewers:
        if hasattr(viewer.window, '_qt_window'):
            with patch.object(
                viewer.window._qt_window, '_save_current_window_settings'
            ):
                viewer.close()
        else:
            viewer.close()

    if GCPASS % 50 == 0 or len(QtViewer._instances):
        gc.collect()
    else:
        gc.collect(1)

    if request.config.getoption(_SAVE_GRAPH_OPNAME):
        fail_obj_graph(QtViewer)

    if request.node.rep_call.failed:
        # IF test failed do not check for leaks
        QtViewer._instances.clear()

    _do_not_inline_below = len(QtViewer._instances)

    QtViewer._instances.clear()  # clear to prevent fail of next test

    # do not inline to avoid pytest trying to compute repr of expression.
    # it fails if C++ object gone but not Python object.
    assert _do_not_inline_below == 0, (
        f'{request.config.getoption(_SAVE_GRAPH_OPNAME)}, {_SAVE_GRAPH_OPNAME}'
    )

    # only check for leaked widgets if an exception was raised during the test,
    # and "strict" mode was used.
    if _strict and getattr(sys, 'last_value', None) is prior_exception:
        QApplication.processEvents()
        leak = set(QApplication.topLevelWidgets()).difference(initial)
        leak = (x for x in leak if x.objectName() != 'handled_widget')
        # still not sure how to clean up some of the remaining vispy
        # vispy.app.backends._qt.CanvasBackendDesktop widgets...
        # observed in `test_sys_info.py`
        if any(n.__class__.__name__ != 'CanvasBackendDesktop' for n in leak):
            # just a warning... but this can be converted to test errors
            # in pytest with `-W error`
            msg = f"""The following Widgets leaked!: {leak}.

            Note: If other tests are failing it is likely that widgets will leak
            as they will be (indirectly) attached to the tracebacks of previous failures.
            Please only consider this an error if all other tests are passing.
            """
            # Explanation notes on the above: While we are indeed looking at the
            # difference in sets of widgets between before and after, new object can
            # still not be garbage collected because of it.
            # in particular with VisPyCanvas, it looks like if a traceback keeps
            # contains the type, then instances are still attached to the type.
            # I'm not too sure why this is the case though.
            if _strict == 'raise':
                raise AssertionError(msg)
            else:
                warnings.warn(msg)


@pytest.fixture
def make_napari_viewer_proxy(make_napari_viewer, monkeypatch):
    """Fixture that returns a function for creating a napari viewer wrapped in proxy.
    Use in the same way like `make_napari_viewer` fixture.

    Parameters
    ----------
    make_napari_viewer : fixture
        The make_napari_viewer fixture.

    Returns
    -------
    function
        A function that creates a napari viewer.
    """
    from napari.utils._proxies import PublicOnlyProxy

    proxies = []

    def actual_factory(*model_args, ensure_main_thread=True, **model_kwargs):
        monkeypatch.setenv(
            'NAPARI_ENSURE_PLUGIN_MAIN_THREAD', str(ensure_main_thread)
        )
        viewer = make_napari_viewer(*model_args, **model_kwargs)
        proxies.append(PublicOnlyProxy(viewer))
        return proxies[-1]

    proxies.clear()

    return actual_factory


@pytest.fixture
def MouseEvent():
    """Create a subclass for simulating vispy mouse events.

    Returns
    -------
    Event : Type
        A new dataclass named Event that can be used to create an
        object with fields "type" and "is_dragging".
    """

    @dataclass
    class Event:
        type: str
        position: tuple[float]
        is_dragging: bool = False
        dims_displayed: tuple[int] = (0, 1)
        dims_point: list[float] = None
        view_direction: list[int] = None
        pos: list[int] = (0, 0)
        button: int = None
        handled: bool = False

    return Event



from collections.abc import Sequence
from numbers import Integral
from typing import (
    Any,
    Literal,
    NamedTuple,
)

import numpy as np

from napari._pydantic_compat import root_validator, validator
from napari.utils.events import EventedModel
from napari.utils.misc import argsort, reorder_after_dim_reduction
from napari.utils.translations import trans


class RangeTuple(NamedTuple):
    start: float
    stop: float
    step: float


class Dims(EventedModel):
    """Dimensions object modeling slicing and displaying.

    Parameters
    ----------
    ndim : int
        Number of dimensions.
    ndisplay : int
        Number of displayed dimensions.
    range : tuple of 3-tuple of float
        List of tuples (min, max, step), one for each dimension in world
        coordinates space. Lower and upper bounds are inclusive.
    point : tuple of floats
        Dims position in world coordinates for each dimension.
    margin_left : tuple of floats
        Left margin in world pixels of the slice for each dimension.
    margin_right : tuple of floats
        Right margin in world pixels of the slice for each dimension.
    order : tuple of int
        Tuple of ordering the dimensions, where the last dimensions are rendered.
    axis_labels : tuple of str
        Tuple of labels for each dimension.
    last_used : int
        Dimension which was last interacted with.

    Attributes
    ----------
    ndim : int
        Number of dimensions.
    ndisplay : int
        Number of displayed dimensions.
    range : tuple of 3-tuple of float
        List of tuples (min, max, step), one for each dimension in world
        coordinates space. Lower and upper bounds are inclusive.
    point : tuple of floats
        Dims position in world coordinates for each dimension.
    margin_left : tuple of floats
        Left margin (=thickness) in world pixels of the slice for each dimension.
    margin_right : tuple of floats
        Right margin (=thickness) in world pixels of the slice for each dimension.
    order : tuple of int
        Tuple of ordering the dimensions, where the last dimensions are rendered.
    axis_labels : tuple of str
        Tuple of labels for each dimension.
    last_used : int
        Dimension which was last used.
        Tuple the slider position for each dims slider, in world coordinates.
    current_step : tuple of int
        Current step for each dimension (same as point, but in slider coordinates).
    nsteps : tuple of int
        Number of steps available to each slider. These are calculated from
        the ``range``.
    thickness : tuple of floats
        Thickness of the slice (sum of both margins) for each dimension in world coordinates.
    displayed : tuple of int
        List of dimensions that are displayed. These are calculated from the
        ``order`` and ``ndisplay``.
    not_displayed : tuple of int
        List of dimensions that are not displayed. These are calculated from the
        ``order`` and ``ndisplay``.
    displayed_order : tuple of int
        Order of only displayed dimensions. These are calculated from the
        ``displayed`` dimensions.
    rollable :  tuple of bool
        Tuple of axis roll state. If True the axis is rollable.
    """

    # fields
    ndim: int = 2
    ndisplay: Literal[2, 3] = 2

    order: tuple[int, ...] = ()
    axis_labels: tuple[str, ...] = ()
    rollable: tuple[bool, ...] = ()

    range: tuple[RangeTuple, ...] = ()
    margin_left: tuple[float, ...] = ()
    margin_right: tuple[float, ...] = ()
    point: tuple[float, ...] = ()

    last_used: int = 0

    # private vars
    _play_ready: bool = True  # False if currently awaiting a draw event
    _scroll_progress: int = 0

    # validators
    # check fields is false to allow private fields to work
    @validator(
        'order',
        'axis_labels',
        'rollable',
        'point',
        'margin_left',
        'margin_right',
        pre=True,
        allow_reuse=True,
    )
    def _as_tuple(v):
        return tuple(v)

    @validator('range', pre=True)
    def _check_ranges(ranges):
        """
        Ensure the range values are sane.

        - start < stop
        - step > 0
        """
        for axis, (start, stop, step) in enumerate(ranges):
            if start > stop:
                raise ValueError(
                    trans._(
                        'start and stop must be strictly increasing, but got ({start}, {stop}) for axis {axis}',
                        deferred=True,
                        start=start,
                        stop=stop,
                        axis=axis,
                    )
                )
            if step <= 0:
                raise ValueError(
                    trans._(
                        'step must be strictly positive, but got {step} for axis {axis}.',
                        deferred=True,
                        step=step,
                        axis=axis,
                    )
                )
        return ranges

    @root_validator(skip_on_failure=True, allow_reuse=True)
    def _check_dims(cls, values):
        """Check the consistency of dimensionality for all attributes.

        Parameters
        ----------
        values : dict
            Values dictionary to update dims model with.
        """
        updated = {}

        ndim = values['ndim']

        range_ = ensure_len(values['range'], ndim, pad_width=(0.0, 2.0, 1.0))
        updated['range'] = tuple(RangeTuple(*rng) for rng in range_)

        point = ensure_len(values['point'], ndim, pad_width=0.0)
        # ensure point is limited to range
        updated['point'] = tuple(
            np.clip(pt, rng.start, rng.stop)
            for pt, rng in zip(point, updated['range'], strict=False)
        )

        updated['margin_left'] = ensure_len(
            values['margin_left'], ndim, pad_width=0.0
        )
        updated['margin_right'] = ensure_len(
            values['margin_right'], ndim, pad_width=0.0
        )

        # order and label default computation is too different to include in ensure_len()
        # Check the order tuple has same number of elements as ndim
        order = values['order']
        if len(order) < ndim:
            order_ndim = len(order)
            # new dims are always prepended
            prepended_dims = tuple(range(ndim - order_ndim))
            # maintain existing order, but shift accordingly
            existing_order = tuple(o + ndim - order_ndim for o in order)
            order = prepended_dims + existing_order
        elif len(order) > ndim:
            order = reorder_after_dim_reduction(order[-ndim:])
        updated['order'] = order

        # Check the order is a permutation of 0, ..., ndim - 1
        if set(updated['order']) != set(range(ndim)):
            raise ValueError(
                trans._(
                    'Invalid ordering {order} for {ndim} dimensions',
                    deferred=True,
                    order=updated['order'],
                    ndim=ndim,
                )
            )

        # Check the axis labels tuple has same number of elements as ndim
        axis_labels = values['axis_labels']
        labels_ndim = len(axis_labels)
        if labels_ndim < ndim:
            # Append new "default" labels to existing ones
            if axis_labels == tuple(map(str, range(labels_ndim))):
                updated['axis_labels'] = tuple(map(str, range(ndim)))
            else:
                updated['axis_labels'] = (
                    tuple(map(str, range(ndim - labels_ndim))) + axis_labels
                )
        elif labels_ndim > ndim:
            updated['axis_labels'] = axis_labels[-ndim:]

        # Check the rollable axes tuple has same number of elements as ndim
        updated['rollable'] = ensure_len(values['rollable'], ndim, True)

        # If the last used slider is no longer visible, use the first.
        last_used = values['last_used']
        ndisplay = values['ndisplay']
        dims_range = updated['range']
        nsteps = cls._nsteps_from_range(dims_range)
        not_displayed = [
            d for d in order[:-ndisplay] if len(nsteps) > d and nsteps[d] > 1
        ]
        if len(not_displayed) > 0 and last_used not in not_displayed:
            updated['last_used'] = not_displayed[0]

        return {**values, **updated}

    @staticmethod
    def _nsteps_from_range(dims_range) -> tuple[float, ...]:
        return tuple(
            # "or 1" ensures degenerate dimension works
            int((rng.stop - rng.start) / (rng.step or 1)) + 1
            for rng in dims_range
        )

    @property
    def nsteps(self) -> tuple[float, ...]:
        return self._nsteps_from_range(self.range)

    @nsteps.setter
    def nsteps(self, value):
        self.range = tuple(
            RangeTuple(
                rng.start, rng.stop, (rng.stop - rng.start) / (nsteps - 1)
            )
            for rng, nsteps in zip(self.range, value, strict=False)
        )

    @property
    def current_step(self):
        return tuple(
            round((point - rng.start) / (rng.step or 1))
            for point, rng in zip(self.point, self.range, strict=False)
        )

    @current_step.setter
    def current_step(self, value):
        self.point = tuple(
            rng.start + point * rng.step
            for point, rng in zip(value, self.range, strict=False)
        )

    @property
    def thickness(self) -> tuple[float, ...]:
        return tuple(
            left + right
            for left, right in zip(
                self.margin_left, self.margin_right, strict=False
            )
        )

    @thickness.setter
    def thickness(self, value):
        self.margin_left = self.margin_right = tuple(val / 2 for val in value)

    @property
    def displayed(self) -> tuple[int, ...]:
        """Tuple: Dimensions that are displayed."""
        return self.order[-self.ndisplay :]

    @property
    def not_displayed(self) -> tuple[int, ...]:
        """Tuple: Dimensions that are not displayed."""
        return self.order[: -self.ndisplay]

    @property
    def displayed_order(self) -> tuple[int, ...]:
        return tuple(argsort(self.displayed))

    def set_range(
        self,
        axis: int | Sequence[int],
        _range: Sequence[int | float] | Sequence[Sequence[int | float]],
    ):
        """Sets ranges (min, max, step) for the given dimensions.

        Parameters
        ----------
        axis : int or sequence of int
            Dimension index or a sequence of axes whos range will be set.
        _range : tuple or sequence of tuple
            Range specified as (min, max, step) or a sequence of these range
            tuples.
        """
        axis, value = self._sanitize_input(
            axis, _range, value_is_sequence=True
        )
        full_range = list(self.range)
        for ax, val in zip(axis, value, strict=False):
            full_range[ax] = val
        self.range = tuple(full_range)

    def set_point(
        self,
        axis: int | Sequence[int],
        value: float | Sequence[float],
    ):
        """Sets point to slice dimension in world coordinates.

        Parameters
        ----------
        axis : int or sequence of int
            Dimension index or a sequence of axes whos point will be set.
        value : scalar or sequence of scalars
            Value of the point for each axis.
        """
        axis, value = self._sanitize_input(
            axis, value, value_is_sequence=False
        )
        full_point = list(self.point)
        for ax, val in zip(axis, value, strict=False):
            full_point[ax] = val
        self.point = tuple(full_point)

    def set_current_step(
        self,
        axis: int | Sequence[int],
        value: int | Sequence[int],
    ):
        axis, value = self._sanitize_input(
            axis, value, value_is_sequence=False
        )
        range_ = list(self.range)
        value_world = []
        for ax, val in zip(axis, value, strict=False):
            rng = range_[ax]
            value_world.append(rng.start + val * rng.step)
        self.set_point(axis, value_world)

    def set_axis_label(
        self,
        axis: int | Sequence[int],
        label: str | Sequence[str],
    ):
        """Sets new axis labels for the given axes.

        Parameters
        ----------
        axis : int or sequence of int
            Dimension index or a sequence of axes whos labels will be set.
        label : str or sequence of str
            Given labels for the specified axes.
        """
        axis, label = self._sanitize_input(
            axis, label, value_is_sequence=False
        )
        full_axis_labels = list(self.axis_labels)
        for ax, val in zip(axis, label, strict=False):
            full_axis_labels[ax] = val
        self.axis_labels = tuple(full_axis_labels)

    def reset(self):
        """Reset dims values to initial states."""
        # Don't reset axis labels
        # TODO: could be optimized with self.update, but need to fix
        #       event firing in EventedModel first
        self.range = ((0, 2, 1),) * self.ndim
        self.point = (0,) * self.ndim
        self.order = tuple(range(self.ndim))
        self.margin_left = (0,) * self.ndim
        self.margin_right = (0,) * self.ndim
        self.rollable = (True,) * self.ndim

    def transpose(self):
        """Transpose displayed dimensions.

        This swaps the order of the last two displayed dimensions.
        The order of the displayed is taken from Dims.order.
        """
        order = list(self.order)
        order[-2], order[-1] = order[-1], order[-2]
        self.order = order

    def _increment_dims_right(self, axis: int | None = None):
        """Increment dimensions to the right along given axis, or last used axis if None

        Parameters
        ----------
        axis : int, optional
            Axis along which to increment dims, by default None
        """
        if axis is None:
            axis = self.last_used
        self.set_current_step(axis, self.current_step[axis] + 1)

    def _increment_dims_left(self, axis: int | None = None):
        """Increment dimensions to the left along given axis, or last used axis if None

        Parameters
        ----------
        axis : int, optional
            Axis along which to increment dims, by default None
        """
        if axis is None:
            axis = self.last_used
        self.set_current_step(axis, self.current_step[axis] - 1)

    def _focus_up(self):
        """Shift focused dimension slider to be the next slider above."""
        sliders = [d for d in self.not_displayed if self.nsteps[d] > 1]
        if len(sliders) == 0:
            return

        index = (sliders.index(self.last_used) + 1) % len(sliders)
        self.last_used = sliders[index]

    def _focus_down(self):
        """Shift focused dimension slider to be the next slider bellow."""
        sliders = [d for d in self.not_displayed if self.nsteps[d] > 1]
        if len(sliders) == 0:
            return

        index = (sliders.index(self.last_used) - 1) % len(sliders)
        self.last_used = sliders[index]

    def roll(self):
        """Roll order of dimensions for display."""
        order = np.array(self.order)
        # we combine "rollable" and "nsteps" into a mask for rolling
        # this mask has to be aligned to "order" as "rollable" and
        # "nsteps" are static but order is dynamic, meaning "rollable"
        # and "nsteps" encode the axes by position, whereas "order"
        # encodes axis by number
        valid = np.logical_and(self.rollable, np.array(self.nsteps) > 1)[order]
        order[valid] = np.roll(order[valid], shift=1)
        self.order = order

    def _go_to_center_step(self):
        self.current_step = [int((ns - 1) / 2) for ns in self.nsteps]

    def _sanitize_input(
        self, axis, value, value_is_sequence=False
    ) -> tuple[list[int], list]:
        """
        Ensure that axis and value are the same length, that axes are not
        out of bounds, and coerces to lists for easier processing.
        """
        if isinstance(axis, Integral):
            if (
                isinstance(value, Sequence)
                and not isinstance(value, str)
                and not value_is_sequence
            ):
                raise ValueError(
                    trans._('cannot set multiple values to a single axis')
                )
            axis = [axis]
            value = [value]
        else:
            axis = list(axis)
            value = list(value)

        if len(axis) != len(value):
            raise ValueError(
                trans._('axis and value sequences must have equal length')
            )

        for ax in axis:
            ensure_axis_in_bounds(ax, self.ndim)
        return axis, value


def ensure_len(value: tuple, length: int, pad_width: Any):
    """
    Ensure that the value has the required number of elements.

    Right-crop if value is too long; left-pad with default if too short.

    Parameters
    ----------
    value : Tuple
        A tuple of values to be resized.
    ndim : int
        Number of desired values.
    default : Tuple
        Default element for left-padding.
    """
    if len(value) < length:
        # left pad
        value = (pad_width,) * (length - len(value)) + value
    elif len(value) > length:
        # right-crop
        value = value[-length:]
    return value


def ensure_axis_in_bounds(axis: int, ndim: int) -> int:
    """Ensure a given value is inside the existing axes of the image.

    Returns
    -------
    axis : int
        The axis which was checked for validity.
    ndim : int
        The dimensionality of the layer.

    Raises
    ------
    ValueError
        The given axis index is out of bounds.
    """
    if axis not in range(-ndim, ndim):
        msg = trans._(
            'Axis {axis} not defined for dimensionality {ndim}. Must be in [{ndim_lower}, {ndim}).',
            deferred=True,
            axis=axis,
            ndim=ndim,
            ndim_lower=-ndim,
        )
        raise ValueError(msg)

    return axis % ndim
